# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import pytest
import unittest
from datetime import datetime

import numpy as np
import numpy.testing as nptest

from ascat.h_saf import H14GribFileList
from ascat.h_saf import AscatNrtBufrFileList
from ascat.h_saf import AscatSsmDataRecord

from get_path import get_testdata_path


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_AscatNrtSsm(unittest.TestCase):

    def setUp(self):
        """
        Setup test data.
        """
        self.root_path = get_testdata_path() / "hsaf"

    def test_h16_read(self):
        """
        Test read file.
        """
        dt = datetime(2017, 2, 20, 11, 15, 0)
        path = self.root_path / 'h16'
        product = 'h16'

        h16 = AscatNrtBufrFileList(path, product)
        data, metadata = h16.read(dt)

        sm_should = np.array([
            0., 3.6, 7.8, 8.2, 12.3, 14.7, 21.6, 26.7, 30.6, 32.2, 43., 50.5,
            46.3, 47.6, 58.
        ])

        lats_should = np.array([
            -28.25222,
            -28.21579,
            -28.1789,
            -28.14155,
            -28.10374,
            -28.06547,
            -28.02674,
            -27.98755,
            -27.94791,
            -27.90782,
            -27.86727,
            -27.82627,
            -27.78482,
            -27.74292,
            -27.70058,
        ])

        sm_mean_should = np.array([
            32.8, 35.1, 37.8, 39.2, 39., 38., 36.2, 39.6, 43.8, 45.6, 46.1,
            47., 43.7, 46., 46.7
        ])

        nptest.assert_allclose(data['lat'][798:813], lats_should, atol=1e-5)
        nptest.assert_allclose(data['sm'][798:813], sm_should, atol=0.01)
        nptest.assert_allclose(
            data['sm_mean'][798:813], sm_mean_should, atol=0.01)

    def test_h101_read(self):
        """
        Test read file.
        """
        dt = datetime(2017, 2, 20, 10, 42, 0)
        path = self.root_path / 'h101'
        product_id = 'h101'

        h101 = AscatNrtBufrFileList(path, product_id)
        data, metadata = h101.read(dt)

        sm_should = np.array([
            26.1, 31.5, 49.5, 64.3, 80.3, 87.9, 24.5, 18.5, 20., 12.9, 6.3,
            3.1, 3.5, 5.7, 2., 0., 4.6, 8., 9.6, 11.8
        ])

        lats_should = np.array([
            56.26292, 56.29267, 56.32112, 56.34827, 56.37413, 56.39868,
            51.98611, 52.08869, 52.19038, 52.29117, 52.39104, 52.48999,
            52.58802, 52.68512, 52.78127, 52.87648, 52.97073, 53.06402,
            53.15634, 53.24768
        ])

        sm_mean_should = np.array([
            22.8, 27., 31.8, 38.3, 44.3, 52.4, 24.8, 27.8, 24.7, 23.9, 22.8,
            25., 25.6, 26.1, 25.9, 26.9, 28.5, 27.8, 24.3, 22.7
        ])

        nptest.assert_allclose(data['lat'][1800:1820], lats_should, atol=1e-5)
        nptest.assert_allclose(data['sm'][1800:1820], sm_should, atol=0.01)
        nptest.assert_allclose(
            data['sm_mean'][1800:1820], sm_mean_should, atol=0.01)

    def test_h102_read(self):
        """
        Test read file.
        """
        dt = datetime(2017, 2, 20, 10, 42, 0)
        path = self.root_path / 'h102'
        product_id = 'h102'

        h102 = AscatNrtBufrFileList(path, product_id)
        data, metadata = h102.read(dt)

        sm_should = np.array([
            45.8, 43.5, 41.7, 42.7, 38.6, 31.1, 23.4, 21.7, 23.6, 26.8, 30.5,
            30.1, 32.7, 34.7, 35.8, 38.3, 46.2, 53.7, 46.2
        ])

        lats_should = np.array([
            43.16844, 43.21142, 43.25423, 43.29686, 43.33931, 43.38158,
            43.42367, 43.46558, 43.50731, 43.54887, 43.59024, 43.63143,
            43.67243, 43.71326, 43.7539, 43.79436, 43.83463, 43.87472, 43.91463
        ])

        sm_mean_should = np.array([
            50.8, 44.5, 36.3, 29.9, 28.7, 29.1, 30., 30.1, 30.8, 33., 34.9,
            35.7, 35.2, 34.2, 32.9, 32.9, 34.5, 32.7, 30.6
        ])

        nptest.assert_allclose(data['lat'][0:19], lats_should, atol=1e-5)
        nptest.assert_allclose(data['sm'][0:19], sm_should, atol=0.01)
        nptest.assert_allclose(
            data['sm_mean'][0:19], sm_mean_should, atol=0.01)

    def test_h103_read(self):
        """
        Test read file.
        """
        dt = datetime(2017, 2, 20, 10, 30, 0)
        path = self.root_path / 'h103'
        product_id = 'h103'

        h103 = AscatNrtBufrFileList(path, product_id)
        data, metadata = h103.read(dt)

        sm_should = np.array([
            20.4, 15.2, 4.2, 0., 0.7, 6.8, 14.2, 21.9, 23.7, 17.4, 14.8, 19.8,
            15.4, 17.1, 28.9, 28.5, 23.9, 18.1, 25.4
        ])

        lats_should = np.array([
            8.82896, 8.85636, 8.88373, 8.91107, 8.93837, 8.96564, 8.99288,
            9.02009, 9.04726, 9.0744, 9.1015, 9.12858, 9.15561, 9.18262,
            9.20959, 9.23653, 9.26343, 9.2903, 9.31713
        ])

        sm_mean_should = np.array([
            14., 14., 13.8, 13.5, 13.3, 13.1, 12.8, 12.6, 12.6, 12., 12.1,
            12.6, 13.2, 14.1, 13.9, 12.4, 10.9, 10.3, 10.7
        ])

        nptest.assert_allclose(data['lat'][0:19], lats_should, atol=1e-5)
        nptest.assert_allclose(data['sm'][0:19], sm_should, atol=0.01)
        nptest.assert_allclose(
            data['sm_mean'][0:19], sm_mean_should, atol=0.01)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_H14(unittest.TestCase):

    def setUp(self):
        """
        Setup test data.
        """
        self.root_path = get_testdata_path() / "hsaf" / "h14"

    def test_read(self):
        """
        Test read file.
        """
        dt = datetime(2014, 5, 15, 0)
        h14 = H14GribFileList(self.root_path)
        data = h14.read(dt)

        assert sorted(data.keys()) == sorted([
            'lat', 'lon', 'SM_layer1_0-7cm', 'SM_layer2_7-28cm',
            'SM_layer3_28-100cm', 'SM_layer4_100-289cm'
        ])

        for var in data:
            assert data[var].shape == (800, 1600)


class Test_AscatSsmDataRecord(unittest.TestCase):

    def setUp(self):
        path = get_testdata_path()

        self.gpi = 3066159
        self.cdr_path = path / "hsaf"
        self.grid_path = str(path / "hsaf" / "grid")
        self.static_layer_path = str(path / "hsaf" / "static_layer")

    def test_read_h25(self):
        """
        Test read H25.
        """
        self.h25 = AscatSsmDataRecord(
            str(self.cdr_path / 'h25'),
            self.grid_path,
            static_layer_path=self.static_layer_path)

        data = self.h25.read(self.gpi, absolute_sm=True)
        assert data.attrs['gpi'] == self.gpi

        np.testing.assert_approx_equal(
            data.attrs['lon'], 19.03533, significant=4)
        np.testing.assert_approx_equal(
            data.attrs['lat'], 70.05438, significant=4)

        assert len(data) == 7737
        assert data.iloc[15].name.to_pydatetime() == datetime(
            2007, 1, 7, 10, 49, 9, 4)
        assert data.iloc[15]['sm'] == 22
        assert data.iloc[15]['ssf'] == 1
        assert data.iloc[15]['sm_noise'] == 6
        assert data.iloc[15]['frozen_prob'] == 0
        assert data.iloc[15]['snow_prob'] == 127
        assert data.iloc[15]['orbit_dir'].decode('utf-8') == 'D'
        assert data.iloc[15]['proc_flag'] == 0

        np.testing.assert_equal(data.iloc[15]['abs_sm_gldas'], np.nan)
        np.testing.assert_approx_equal(data.iloc[15]['abs_sm_noise_gldas'],
                                       np.nan)

        np.testing.assert_approx_equal(
            data.iloc[15]['abs_sm_hwsd'], 0.1078, significant=6)
        np.testing.assert_approx_equal(
            data.iloc[15]['abs_sm_noise_hwsd'], 0.0294, significant=6)

        assert data.attrs['topo_complex'] == 9
        assert data.attrs['wetland_frac'] == 41

        np.testing.assert_approx_equal(
            data.attrs['porosity_gldas'], np.nan, significant=5)
        np.testing.assert_approx_equal(
            data.attrs['porosity_hwsd'], 0.49000001, significant=5)

    def test_read_h108(self):
        """
        Test read H108.
        """
        self.h108 = AscatSsmDataRecord(
            str(self.cdr_path / 'h108'),
            self.grid_path,
            static_layer_path=self.static_layer_path)

        data = self.h108.read(self.gpi, absolute_sm=True)
        assert data.attrs['gpi'] == self.gpi

        np.testing.assert_approx_equal(
            data.attrs['lon'], 19.03533, significant=4)
        np.testing.assert_approx_equal(
            data.attrs['lat'], 70.05438, significant=4)

        assert len(data) == 8222
        assert data.iloc[15].name.to_pydatetime() == datetime(
            2007, 1, 7, 10, 49, 9, 4)
        assert data.iloc[15]['sm'] == 22
        assert data.iloc[15]['ssf'] == 2
        assert data.iloc[15]['sm_noise'] == 6
        assert data.iloc[15]['frozen_prob'] == 0
        assert data.iloc[15]['snow_prob'] == 127
        assert data.iloc[15]['orbit_dir'].decode('utf-8') == 'D'
        assert data.iloc[15]['proc_flag'] == 0

        np.testing.assert_equal(data.iloc[15]['abs_sm_gldas'], np.nan)
        np.testing.assert_approx_equal(data.iloc[15]['abs_sm_noise_gldas'],
                                       np.nan)

        np.testing.assert_approx_equal(
            data.iloc[15]['abs_sm_hwsd'], 0.1078, significant=6)
        np.testing.assert_approx_equal(
            data.iloc[15]['abs_sm_noise_hwsd'], 0.0294, significant=6)

        assert data.attrs['topo_complex'] == 9
        assert data.attrs['wetland_frac'] == 41

        np.testing.assert_equal(data.attrs['porosity_gldas'], np.nan)
        np.testing.assert_approx_equal(
            data.attrs['porosity_hwsd'], 0.49000001, significant=5)

    def test_read_h109(self):
        """
        Test read H109.
        """
        self.h109 = AscatSsmDataRecord(
            str(self.cdr_path / 'h109'),
            self.grid_path,
            static_layer_path=self.static_layer_path)

        data = self.h109.read(self.gpi, absolute_sm=True)
        assert data.attrs['gpi'] == self.gpi

        np.testing.assert_approx_equal(
            data.attrs['lon'], 19.03533, significant=4)
        np.testing.assert_approx_equal(
            data.attrs['lat'], 70.05438, significant=4)

        assert len(data) == 11736
        assert data.iloc[15].name.to_pydatetime() == \
            datetime(2007, 1, 7, 10, 49, 9, 379200)
        assert data.iloc[15]['sm'] == 27
        assert data.iloc[15]['ssf'] == 1
        assert data.iloc[15]['sm_noise'] == 5
        assert data.iloc[15]['frozen_prob'] == 0
        assert data.iloc[15]['snow_prob'] == 127
        assert data.iloc[15]['dir'] == 1
        assert data.iloc[15]['proc_flag'] == 0
        assert data.iloc[15]['corr_flag'] == 16
        assert data.iloc[15]['sat_id'] == 3

        np.testing.assert_equal(data.iloc[15]['abs_sm_gldas'], np.nan)
        np.testing.assert_approx_equal(data.iloc[15]['abs_sm_noise_gldas'],
                                       np.nan)

        np.testing.assert_approx_equal(
            data.iloc[15]['abs_sm_hwsd'], 0.1323, significant=6)
        np.testing.assert_approx_equal(
            data.iloc[15]['abs_sm_noise_hwsd'], 0.0245, significant=6)

        assert data.attrs['topo_complex'] == 9
        assert data.attrs['wetland_frac'] == 41

        np.testing.assert_equal(data.attrs['porosity_gldas'], np.nan)
        np.testing.assert_approx_equal(
            data.attrs['porosity_hwsd'], 0.49000001, significant=5)

    def test_read_h110(self):
        """
        Test read H110.
        """
        self.h110 = AscatSsmDataRecord(
            str(self.cdr_path / 'h110'),
            self.grid_path,
            static_layer_path=self.static_layer_path)

        data = self.h110.read(self.gpi, absolute_sm=True)
        assert data.attrs['gpi'] == self.gpi

        np.testing.assert_approx_equal(
            data.attrs['lon'], 19.03533, significant=4)
        np.testing.assert_approx_equal(
            data.attrs['lat'], 70.05438, significant=4)

        assert len(data) == 1148
        assert data.iloc[15].name.to_pydatetime() == datetime(
            2016, 1, 3, 19, 34, 28, 99200)
        assert data.iloc[15]['sm'] == 48
        assert data.iloc[15]['ssf'] == 1
        assert data.iloc[15]['sm_noise'] == 5
        assert data.iloc[15]['frozen_prob'] == 0
        assert data.iloc[15]['snow_prob'] == 127
        assert data.iloc[15]['dir'] == 0
        assert data.iloc[15]['proc_flag'] == 0
        assert data.iloc[15]['corr_flag'] == 0
        assert data.iloc[15]['sat_id'] == 4

        np.testing.assert_equal(data.iloc[15]['abs_sm_gldas'], np.nan)
        np.testing.assert_approx_equal(data.iloc[15]['abs_sm_noise_gldas'],
                                       np.nan)

        np.testing.assert_approx_equal(
            data.iloc[15]['abs_sm_hwsd'], 0.2352, significant=6)
        np.testing.assert_approx_equal(
            data.iloc[15]['abs_sm_noise_hwsd'], 0.0245, significant=6)

        assert data.attrs['topo_complex'] == 9
        assert data.attrs['wetland_frac'] == 41

        np.testing.assert_equal(data.attrs['porosity_gldas'], np.nan)
        np.testing.assert_approx_equal(
            data.attrs['porosity_hwsd'], 0.49000001, significant=5)

    def test_read_h111(self):
        """
        Test read H111.
        """
        self.h111 = AscatSsmDataRecord(
            str(self.cdr_path / 'h111'),
            self.grid_path,
            static_layer_path=self.static_layer_path)

        data = self.h111.read(self.gpi, absolute_sm=True)
        assert data.attrs['gpi'] == self.gpi

        np.testing.assert_approx_equal(
            data.attrs['lon'], 19.03533, significant=4)
        np.testing.assert_approx_equal(
            data.attrs['lat'], 70.05438, significant=4)

        assert len(data) == 13715
        assert data.iloc[15].name.to_pydatetime() == datetime(
            2007, 1, 7, 10, 49, 9, 379200)
        assert data.iloc[15]['sm'] == 28
        assert data.iloc[15]['ssf'] == 1
        assert data.iloc[15]['sm_noise'] == 5
        assert data.iloc[15]['frozen_prob'] == 0
        assert data.iloc[15]['snow_prob'] == 127
        assert data.iloc[15]['dir'] == 1
        assert data.iloc[15]['proc_flag'] == 0
        assert data.iloc[15]['corr_flag'] == 4
        assert data.iloc[15]['sat_id'] == 3

        np.testing.assert_equal(data.iloc[15]['abs_sm_gldas'], np.nan)
        np.testing.assert_approx_equal(data.iloc[15]['abs_sm_noise_gldas'],
                                       np.nan)

        np.testing.assert_approx_equal(
            data.iloc[15]['abs_sm_hwsd'], 0.1372, significant=6)
        np.testing.assert_approx_equal(
            data.iloc[15]['abs_sm_noise_hwsd'], 0.0245, significant=6)

        assert data.attrs['topo_complex'] == 9
        assert data.attrs['wetland_frac'] == 41

        np.testing.assert_equal(data.attrs['porosity_gldas'], np.nan)
        np.testing.assert_approx_equal(
            data.attrs['porosity_hwsd'], 0.49000001, significant=5)

    def test_read_2points_cell_switch(self):
        """
        Test reading of two points in two different cells.
        This did not work in the past when the static layer class
        was closed too soon.
        """
        self.h111 = AscatSsmDataRecord(
            str(self.cdr_path / 'h111'),
            self.grid_path,
            static_layer_path=self.static_layer_path)

        gpi = 3066159
        data = self.h111.read(gpi, absolute_sm=True)
        assert data.attrs['gpi'] == gpi

        gpi = 2577735
        data = self.h111.read(gpi, absolute_sm=True)
        assert data.attrs['gpi'] == gpi


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2025, TU Wien
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import tempfile
from pathlib import Path

import numpy as np
import xarray as xr

from ascat.cf_array import indexed_to_contiguous
from ascat.cf_array import contiguous_to_indexed

dtype_to_nan = {
    np.dtype("int8"): -2**7,
    np.dtype("uint8"): 2**8 - 1,
    np.dtype("int16"): -2**15,
    np.dtype("uint16"): 2**16 - 1,
    np.dtype("int32"): -2**31,
    np.dtype("uint32"): 2**32 - 1,
    np.dtype("int64"): -2**63,
    np.dtype("uint64"): 2**64 - 1,
    np.dtype("float32"): -2**31,
    np.dtype("float64"): -2**31,
    np.dtype("<M8[ns]"): 0,
    np.dtype("<M8[s]"): 0,
}


def verify_ortho_multi(ds: xr.Dataset, instance_dim: str,
                       element_dim: str) -> None:
    """
    Verify dataset follows orthogonal multidimensional array CF definition.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to be verified.
    instance_dim : str
        Name of the instance dimension.
    element_dim : str
        Name of the element dimension.

    Returns
    -------
    sample_dimension : str
        Name of the sample dimension.

    Raises
    ------
    RuntimeError if verification fails.
    """
    # check that instance dimension exists
    if instance_dim not in ds.dims:
        raise RuntimeError(f"Instance dimension is missing '{instance_dim}'")

    # check that element dimension exists
    if element_dim not in ds.dims:
        raise RuntimeError(f"Element dimension is missing '{element_dim}'")


def verify_contiguous_ragged(ds: xr.Dataset, count_var: str,
                             instance_dim: str) -> None:
    """
    Verify dataset follows contiguous ragged array CF definition.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to be verified.
    count_var : str
        Name of the count variable.
        Count variable contains the length of each time series feature. It is
        identified by having an attribute with name 'sample_dimension' whose
        value is name of the sample dimension. The count variable implicitly
        partitions into individual instances all variables that have the
        sample dimension.

    Raises
    ------
    RuntimeError if verification fails.
    """
    # check that count variable exists
    if count_var not in ds:
        raise RuntimeError(f"Count variable is missing: {count_var}")

    # check that count variable contains sample_dimension attribute
    if "sample_dimension" not in ds[count_var].attrs:
        raise RuntimeError(f"Count variable '{count_var}' has no "
                           "sample_dimension attribute")

    # check that count variable has instance_dimension as single dimension
    if ds[count_var].dims != (instance_dim,):
        raise RuntimeError(f"Count variable '{count_var}' must have the "
                           f"instance dimension '{instance_dim}' as its "
                           "single dimension")


def verify_indexed_ragged(ds: xr.Dataset, index_var: str,
                          sample_dim: str) -> None:
    """
    Verify dataset follows indexed ragged array CF definition.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset.
    index_var : str
        The index variable can be identified by having an attribute with
        name of instance_dimension whose value is the instance dimension.
    sample_dim : str
        Name of the sample dimension.

    Raises
    ------
    RuntimeError if verification fails.
    """
    # check that index variable exists
    if index_var not in ds:
        raise RuntimeError(f"Index variable is missing: {index_var}")

    # check that index variable must have sample dimension as single dimension
    if ds[index_var].dims != (sample_dim,):
        raise RuntimeError(f"Index variable '{index_var}' must have the "
                           f"sample dimension '{sample_dim}' as its "
                           "single dimension")

    # check that index variable has instance_dimension attribute
    if "instance_dimension" not in ds[index_var].attrs:
        raise RuntimeError(f"Index variable '{index_var}' has no "
                           "instance_dimension attribute")


def verify_point_array(ds: xr.Dataset, sample_dim: str) -> None:
    """
    Verify dataset follows the CF point data array convention.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to be verified.
    sample_dim : str
        Name of the sample dimension.

    Raises
    ------
    RuntimeError if verification fails.
    """
    # check that the sample_dim exists
    if sample_dim not in ds.dims:
        raise RuntimeError(f"Sample dimension '{sample_dim}' is missing.")

    # check all data and coordinate variables have only the sample_dim
    for var in ds.variables:
        dims = ds[var].dims
        if ds[var].ndim > 0 and dims != (sample_dim,):
            raise RuntimeError(
                f"Variable '{var}' does not conform to point structure "
                f"(dims: {dims}). All variables must use only the sample_dim ('{sample_dim}')."
            )


class PointData:
    """
    Point data represent scattered locations and times with no implied
    relationship among of coordinate positions, both data and coordinates must
    share the same (sample) instance dimension.
    """

    def __init__(self, ds: xr.Dataset, sample_dim: str):
        """
        Initialize.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset to be verified.
        sample_dim : str
            Name of the sample dimension. The sample dimension indicates the
            number of instances (e.g. stations, locations).
        """
        self.sample_dim = sample_dim
        self._data = ds
        self.validate()

    def validate(self):
        """Validate format."""
        verify_point_array(self.ds, self.sample_dim)

    @property
    def ds(self):
        return self._data

    def to_indexed(self, index_var: str = "obs", instance_dim: str = "loc"):
        """
        Convert point data to indexed ragged array.

        Parameters
        ----------
        index_var : str
            Name of the new index variable to be added.
        instance_dim : str
            Name of the instance dimension.

        Returns
        -------
        indexed : IndexedRaggedArray
            Indexed ragged array object.
        """
        if instance_dim not in self.ds:
            raise ValueError(
                f"'{instance_dim}' must be a coordinate in the dataset")

        instance_ids, inverse_index = np.unique(
            self.ds[instance_dim], return_inverse=True)

        # create index variable (mapping each sample to its instance)
        index_data = xr.DataArray(
            inverse_index,
            dims=(self.sample_dim,),
            attrs={"instance_dimension": instance_dim})

        new_ds = self.ds.copy()
        new_ds[index_var] = index_data
        new_ds = new_ds.assign_coords({instance_dim: instance_ids})

        return IndexedRaggedArray(
            new_ds, index_var=index_var, sample_dim=self.sample_dim)

    def to_contiguous(self,
                      count_var: str = "row_size",
                      instance_dim: str = "loc"):
        """
        Convert point data to contiguous ragged array.

        Parameters
        ----------
        count_var : str
            Name of the new count variable to be added (default: 'row_size').
        instance_dim : str
            Name of the instance dimension (default: 'loc').

        Returns
        -------
        contiguous : ContiguousRaggedArray
            Contiguous ragged array object.
        """
        if instance_dim not in self.ds:
            raise ValueError(
                f"'{instance_dim}' must be a coordinate in the dataset")

        group = self.ds.groupby(self.ds[instance_dim])
        row_sizes = group.groups
        instance_ids = list(row_sizes.keys())
        counts = np.array([len(row_sizes[i]) for i in instance_ids])

        sort_index = np.argsort(self.ds[instance_dim].values)
        sorted_ds = self.ds.isel({self.sample_dim: sort_index})

        count_data = xr.DataArray(
            counts,
            dims=(instance_dim,),
            attrs={"sample_dimension": self.sample_dim})

        new_ds = sorted_ds.copy()
        new_ds[count_var] = count_data

        if instance_dim in new_ds.coords:
            new_ds = new_ds.assign_coords(
                {instance_dim: np.array(instance_ids)})

        return ContiguousRaggedArray(
            new_ds, count_var=count_var, instance_dim=instance_dim)


class OrthoMultiArray:
    """
    Orthogonal multidimensional array.

    Attributes
    ----------
    instance_dim : str
        Name of the instance dimension.
    element_dim : str
        Element dimension name.
    ds : xarray.Dataset
        Orthomulti array dataset.
    instance_variables : list
        List of instance variables.

    Methods
    -------
    sel_instance(i)
        Read time series for given instance.
    iter()
        Yield time series for each instance.
    """

    def __init__(self,
                 ds: xr.Dataset,
                 instance_dim: str = "loc",
                 element_dim: str = "time"):
        """
        Initialize.

        Parameters
        ----------
        ds : xr.Dataset
            Data stored in orthomulti array format.
        instance_dim : str
            Instance dimension name.
        element_dim : str
            Element dimension name.
        """
        self.instance_dim = instance_dim
        self.element_dim = element_dim
        self._data = ds
        self.validate()

        # if instance_dim exists as a variable, it is assumed these are there
        # instance identifiers used as keys/reading
        if self.instance_dim in ds:
            self._instances = ds[self.instance_dim].to_numpy()
            self._instance_lookup = np.zeros(
                self._instances.max() + 1,
                dtype=np.int64) + self._instances.size
            self._instance_lookup[self._instances] = np.arange(
                self._instances.size)
        else:
            self._instances = np.arange(ds.sizes[instance_dim])
            self._instance_lookup = np.arange(self._instances.size)

    def validate(self):
        """Validate format."""
        verify_ortho_multi(self.ds, self.instance_dim, self.element_dim)

    @property
    def ds(self):
        return self._data

    def sel_instance(self, instance_id: int):
        """Read time series"""
        return self.ds.sel({self.instance_dim: instance_id})

    def __iter__(self):
        """Iterator over time series"""
        for instance in self.ds[self.instance_dim]:
            yield self.sel_instance(instance)

    def iter(self):
        """Explicit iterator method"""
        return self.__iter__()

    def to_point_data(self):
        raise NotImplementedError(
            "Conversion to point data not implemented yet.")

    def to_indexed(self):
        raise NotImplementedError(
            "Conversion to indexed ragged not implemented yet.")

    def to_contiguous(self):
        raise NotImplementedError(
            "Conversion to contiguous ragged not implemented yet.")

    def apply(self, func):
        pass


class ContiguousRaggedArray:
    """
    Contiguous ragged array representation (CF convention).

    In an contiguous ragged array representation, the dataset for all time
    series are stored in a single 1D array. Additional variables or
    dimensions provide the metadata needed to map these values back to their
    respective time series.

    The contiguous ragged array representation can be used only if the size
    of each instance is known at the time that it is created. In this
    representation the data for each instance will be contiguous on disk.

    If the instance dimension exists as a variable, it is assumed that the
    values represent the identifiers for each instance otherwise they are
    count upwards from 0.

    Attributes
    ----------
    instance_dim : str
        Name of the instance dimension.
    sample_dim : str
        Name of the sample dimension. The variable bearing the
        sample_dimension attribute (i.e. count_var) must have the instance
        dimension as its single dimension, and must have an integer type.
    count_var : str
        Name of the count variable. The count variable must be an integer
        type and must have the instance dimension as its sole dimension.
        The count variable are identifiable by the presence of an attribute,
        sample_dimension, found on the count variable, which names the sample
        dimension being counted.
    ds : xarray.Dataset
        Contiguous ragged array dataset.
    instance_variables : list
        List of instance variables.
    instance_ids : list
        List of instance ids.

    Methods
    -------
    sel_instance(i)
        Read time series for given instance.
    iter()
        Yield time series for each instance.
    """

    def __init__(self,
                 ds: xr.Dataset,
                 count_var: str,
                 instance_dim: str,
                 instance_id_var: str = None):
        """
        Initialize.

        Parameters
        ----------
        ds : xr.Dataset
            Data stored in contiguous ragged array format.
        count_var : str
            Count variable name.
        instance_dim : str
            Instance dimension name.
        instance_id_var: str, optional
            Variable used as instance identifier (default: None).
        """
        self.count_var = count_var
        self.instance_dim = instance_dim
        self._data = ds
        self.validate()

        self.sample_dim = ds[count_var].attrs["sample_dimension"]
        self.instance_id_var = instance_id_var
        self._lut = None

        # cache row_size and instance_ids data
        self._row_size = self.ds[self.count_var].to_numpy()

        if self.instance_id_var is None:
            self._instance_ids = self.ds[self.instance_dim].to_numpy()
        else:
            self._instance_ids = self.ds[self.instance_id_var].to_numpy()

        self._set_instance_lut()

    def validate(self):
        """Validate format."""
        verify_contiguous_ragged(self.ds, self.count_var, self.instance_dim)

    def _set_instance_lut(self):
        """
        Set instance lookup-table.
        """
        self._instance_lookup = np.zeros(
            self._instance_ids.max() + 1,
            dtype=np.int64) + self._instance_ids.size
        self._instance_lookup[self._instance_ids] = np.arange(
            self._instance_ids.size)

        self._lut = np.zeros(self._instance_ids.max() + 1, dtype=bool)

    @classmethod
    def from_file(cls,
                  filename: str,
                  count_var: str,
                  instance_dim: str,
                  instance_id_var: str = None,
                  **kwargs):
        """
        Load time series from file.

        Parameters
        ----------
        filename : str
            Filename.
        count_var : str
            Count variable name.
        instance_dim : str
            Instance dimension name.
        instance_id_var: str, optional
            Variable used as instance identifier (default: None).

        Returns
        -------
        data : ContiguousRaggedArray
            ContiguousRaggedArray object loaded from a file.
        """
        ds = xr.open_dataset(filename, **kwargs)
        verify_contiguous_ragged(ds, count_var, instance_dim)

        return cls(ds, count_var, instance_dim, instance_id_var)

    @property
    def ds(self):
        """
        Dataset.

        Returns
        -------
        ds : xr.Dataset
            Contiguous ragged array dataset.
        """
        return self._data

    @property
    def size(self) -> list:
        """
        Number of instances.

        Returns
        -------
        instance_ids : int
            Number of instance.
        """
        return self._instance_ids.size

    @property
    def instance_ids(self) -> list:
        """
        Instance ids

        Returns
        -------
        instance_ids : list of int
            Instance ids.
        """
        return self._instance_ids

    @property
    def instance_variables(self) -> list:
        """
        Instance variables.

        Returns
        -------
        instance_variables : list of str
            Instance variables.
        """
        return [
            var for var in self.ds.variables
            if (self.ds[var].dims == (self.sample_dim,)) and
            (var != self.sample_dim)
        ]

    def get_instance_variables(self, include_dtype: bool = False) -> list:
        """
        Instance variables.

        Returns
        -------
        instance_variables : list of str
            Instance variables.
        """
        if include_dtype:
            instance_variables = [
                (var, self.ds[var].dtype)
                for var in self.ds.variables
                if (self.ds[var].dims == (self.sample_dim,)) and
                (var != self.sample_dim)
            ]
        else:
            instance_variables = [
                var for var in self.ds.variables
                if (self.ds[var].dims == (self.sample_dim,)) and
                (var != self.sample_dim)
            ]

        return instance_variables

    def sel_instance(self, i: int):
        """Read time series"""
        try:
            idx = self._instance_lookup[i]
        except IndexError:
            data = None
        else:
            if idx == self._instance_ids.size:
                data = None
            else:
                start = self._row_size[:idx].sum()
                end = start + self._row_size[idx]
                data = self.ds.isel({
                    self.sample_dim: slice(start, end),
                    self.instance_dim: idx
                })

        return data

    def sel_instances(self, i: np.ndarray) -> xr.Dataset:
        """
        Read time series for given instance IDs using a LUT and preserve order.

        Parameters
        ----------
        i : np.ndarray
            Array of instance IDs.

        Returns
        -------
        ds : xr.Dataset
            Dataset containing the selected instances in the correct order.
        """
        i = np.asarray(i)
        idx_array = self._instance_lookup[i]

        # mark selected instances in lookup table
        self._lut[idx_array] = True

        # map each sample index to its instance index
        obs = np.repeat(np.arange(len(self._row_size)), self._row_size)

        data = self.ds.sel({
            self.sample_dim: self._lut[obs],
            self.instance_dim: idx_array,
        })

        # reset lookup table
        self._lut[:] = False

        # sort the selected data to match the order in i
        # assuming instance_dim is coordinate-based and one-dimensional
        _, order = np.unique(i, return_inverse=True)

        data = data.isel({self.instance_dim: order})
        data = data.assign_coords({self.instance_dim: i})

        return data

    def __iter__(self):
        """
        Iterator over instances.

        Returns
        -------
        ds : xr.Dataset
            Time series for instance.
        """
        for i in self.instance_ids:
            yield self.sel_instance(i)

    def iter(self):
        """
        Explicit iterator method.

        Returns
        -------
        ds : xr.Dataset
            Time series for instance.
        """
        return self.__iter__()

    def to_indexed(self):
        """
        Convert to indexed ragged array.

        Returns
        -------
        data : IndexedRaggedArray
            Indexed ragged array time series.
        """
        ds = contiguous_to_indexed(self.ds, self.sample_dim, self.instance_dim,
                                   self.count_var, self.sample_dim)

        return IndexedRaggedArray(ds, self.sample_dim, self.sample_dim)

    def to_orthomulti(self):
        """
        Convert to orthogonal multidimensional array.

        Returns
        -------
        data : OrthoMultiArray
            Orthogonal multidimensional array time series.
        """
        # element length defined by longest time series
        element_len = self._row_size.max()
        instance_len = self.instance_ids.size

        # compute matrix coordinates
        x = np.arange(self._row_size.size).repeat(self._row_size)
        y = vrange(np.zeros_like(self._row_size), self._row_size)
        shape = (instance_len, element_len)

        reshaped_ds = xr.Dataset(
            {
                var: ([self.instance_dim, self.sample_dim
                      ], pad_to_2d(self.ds[var], x, y, shape)
                     ) for var in self.instance_variables
            },
            coords={
                self.instance_dim: self.instance_ids,
            },
        )
        for var in self.instance_variables:
            reshaped_ds[var].encoding["_FillValue"] = dtype_to_nan[
                reshaped_ds[var].dtype]

        return OrthoMultiArray(reshaped_ds, self.instance_dim, self.sample_dim)

    def to_point_data(self):
        raise NotImplementedError(
            "Conversion to point data not implemented yet.")

    def apply(self, func):
        """
        Apply function on each instance.
        """
        return self.ds.groupby(self.sample_dim).map(func)


class IndexedRaggedArray:
    """
    Indexed ragged array representation (CF convention).

    In an indexed ragged array representation, the dataset is structured
    to store variable-length data (e.g., time series with varying lengths)
    compactly. To achieve this, auxiliary indexing variables that map the
    flat array storage to meaningful groups (e.g. locations).

    If the instance dimension exists as a variable, it is assumed that the
    values represent the identfiers for each instance otherwise they counting
    upwards from 0.

    Attributes
    ----------
    index_var : str
        The indexed ragged array representation must contain an index
        variable, which must be an integer type, and must have the sample
        dimension as its single dimension.
        The index variable can be identified by having an attribute
        'instance_dimension' whose value is the instance dimension.
    sample_dim : str
        Name of the sample dimension. The sample dimension indicates the
        number of instances (e.g. stations, locations).
    instance_dim : str
        The name of the instance dimension. The value is defined by
        the 'instance_dimension' attribute, which must be present on the
        index variable. All variables having the instance dimension are
        instance variables, i.e. variables holding time series data.
    ds : xarray.Dataset
        Indexed ragged array dataset.
    instance_variables : list
        List of instance variables.
    instance_ids : list
        List of instance ids.

    Methods
    -------
    sel_instance(i)
        Read time series for given instance.
    iter()
        Yield time series for each instance.
    """

    def __init__(self, ds: xr.Dataset, index_var: str, sample_dim: str):
        """
        Initialize.

        Parameters
        ----------
        ds : xr.Dataset
            Data in indexed ragged array structure.
        index_var : str
            Index variable name.
        sample_dim : str
            Sample dimension name.
        """
        self.index_var = index_var
        self.sample_dim = sample_dim
        self._data = ds.set_coords(self.index_var).set_xindex(self.index_var)
        self.validate()

        self.instance_dim = ds[index_var].attrs["instance_dimension"]
        self._instance_lut = None
        self._lut = None
        self._set_instance_lut()

    def validate(self):
        """Validate format."""
        verify_indexed_ragged(self.ds, self.index_var, self.sample_dim)

    def __repr__(self):
        """"""
        return self.ds.__repr__()

    def _set_instance_lut(self):
        """
        Set instance lookup-table.
        """
        if self.instance_dim in self.ds:
            instance_ids = self.ds[self.instance_dim].to_numpy()
            self._instance_lut = np.zeros(
                instance_ids.max() + 1, dtype=np.int64) - 1
            self._instance_lut[instance_ids] = np.arange(instance_ids.size)
        else:
            instance_ids = np.unique(self.ds[self.index_var])
            self._instance_lut = np.arange(instance_ids.size)

        self._lut = np.zeros(instance_ids.max() + 1, dtype=bool)

    @classmethod
    def from_file(cls, filename: str, index_var: str, sample_dim: str):
        """
        Read data from file.

        Parameters
        ----------
        filename : str
            Filename.
        index_var : str
            Index variable name.
        sample_dim : str
            Sample dimension name.

        Returns
        -------
        data : IndexRaggedArray
            IndexRaggedArray object loaded from a file.
        """
        ds = xr.open_dataset(filename)
        verify_indexed_ragged(ds, index_var, sample_dim)
        return cls(ds, index_var, sample_dim)

    def save(self, filename: str):
        """
        Write data to file.

        Parameters
        ----------
        filename : str
            Filename.
        """
        suffix = Path(filename).suffix

        if suffix == ".nc":
            self.ds.to_netcdf(filename)
        elif suffix == ".zarr":
            self.ds.to_zarr(filename)
        else:
            raise ValueError(f"Unknown file suffix '{suffix}' "
                             "(.nc and .zarr supported)")

    @property
    def ds(self) -> xr.Dataset:
        """
        Dataset.

        Returns
        -------
        ds : xr.Dataset
            Indexed ragged array dataset.
        """
        return self._data

    @property
    def size(self) -> list:
        """
        Number of instances.

        Returns
        -------
        instance_ids : int
            Number of instance.
        """
        return self.instance_ids.size

    @property
    def instance_ids(self) -> list:
        """
        Instance ids.

        Returns
        -------
        instance_ids : list of int
            Instance ids.
        """
        return self.ds[self.instance_dim].values

    @property
    def instance_variables(self) -> list:
        """
        Instance variables.

        Returns
        -------
        instance_variables : list of str
            Instance variables.
        """
        return [
            var for var in self.ds.variables
            if (self.ds[var].dims == (self.sample_dim,)) and
            (var != self.sample_dim)
        ]

    def sel_instance(self, i: int) -> xr.Dataset:
        """
        Read time series.

        Parameters
        ----------
        i : int
            Instance identifier.

        Returns
        -------
        ds : xr.Dataset
            Time series for instance.
        """
        data = self.ds.sel({
            self.index_var: self._instance_lut[i],
            self.instance_dim: i
        })

        # reset index variable or drop index variable (my preference)?
        data[self.index_var] = (self.sample_dim,
                                np.zeros(data[self.index_var].size, dtype=int))

        return data

    def sel_instances(self,
                      i: np.array,
                      ignore_missing: bool = True) -> xr.Dataset:
        """
        Select multiple instances (time series).

        Parameters
        ----------
        i : numpy.array
            Instance identifier.

        Returns
        -------
        ds : xr.Dataset
            Time series for instance.
        """
        # check for duplicates in instance identifier?
        i = np.asarray(i)

        # check if any missing instance ids have been selected
        if ignore_missing:
            valid = np.where(self._instance_lut[i] != -1)[0]
            if valid.size == 0:
                raise ValueError("No valid instances selected")
        else:
            if np.any(self._instance_lut[np.asarray(i)] == -1):
                raise ValueError("Missing instances selected")
            else:
                valid = np.ones(i.size, dtype=bool)

        i = i[valid]

        # initialize look-up table
        self._lut[self._instance_lut[i]] = True

        data = self.ds.sel(
            {self.index_var: self._lut[self.ds[self.index_var]]})

        # reset index variable
        index = data[self.instance_dim][data[self.index_var]].to_numpy()
        lut = np.zeros(index.max() + 1, dtype=np.int64)
        # n_inst = self.ds.sel({self.instance_dim: i}).sizes[self.instance_dim]
        n_inst = i.size
        lut[i] = np.arange(n_inst)
        data[self.index_var] = (self.sample_dim, lut[index])
        data = data.sel({self.instance_dim: i})

        # copy attributes
        data[self.index_var].attrs = self.ds[self.index_var].attrs

        # reset look-up table
        self._lut[:] = False

        return IndexedRaggedArray(data, self.index_var, self.sample_dim)

    def __iter__(self) -> xr.Dataset:
        """
        Iterator over instances.

        Returns
        -------
        ds : xr.Dataset
            Time series for instance.
        """
        for i in self.instance_ids:
            yield self.sel_instance(i)

    def iter(self) -> xr.Dataset:
        """
        Explicit iterator method.

        Returns
        -------
        ds : xr.Dataset
            Time series for instance.
        """
        return self.__iter__()

    def to_contiguous(self,
                      count_var: str = "row_size") -> ContiguousRaggedArray:
        """
        Convert to contiguous ragged array.

        Parameters
        ----------
        count_var : str, optional
            Count variable (default: "row_size").

        Returns
        -------
        data : ContiguousRaggedArray
            Contiguous ragged array time series.
        """
        ds = indexed_to_contiguous(self.ds, self.sample_dim, self.instance_dim,
                                   count_var, self.index_var)

        return ContiguousRaggedArray(ds, self.instance_dim, self.instance_dim,
                                     count_var)

    def to_orthomulti(self) -> OrthoMultiArray:
        """
        Convert to orthogonal multidimensional array.

        Returns
        -------
        data : OrthoMultiArray
            Orthogonal multidimensional array time series.
        """
        ds = self.ds.sortby([self.index_var])
        _, row_size = np.unique(ds[self.index_var], return_counts=True)

        # size defined by longest time series
        element_len = row_size.max()
        instance_len = ds.sizes[self.instance_dim]

        # compute matrix coordinates
        x = np.arange(row_size.size).repeat(row_size)
        y = vrange(np.zeros_like(row_size), row_size)
        shape = (instance_len, element_len)

        reshaped_ds = xr.Dataset(
            {
                var: ([self.instance_dim, self.sample_dim],
                      pad_to_2d(ds[var], x, y, shape)) for var in ds.data_vars
            },
            coords={
                self.instance_dim: ds[self.instance_dim],
            },
        )

        return OrthoMultiArray(reshaped_ds, self.instance_dim, self.sample_dim)

    def to_point_data(self):
        raise NotImplementedError(
            "Conversion to point data not implemented yet.")

    def apply(self, func):
        """
        Apply function on each instance.
        """
        # return self.ds.groupby(self.sample_dim).apply(func)
        return self.ds.groupby(self.sample_dim).map(func)

    def append(self, ds: xr.Dataset):
        """
        Append indexed ragged array time series.

        Parameters
        ----------
        ds : xarray.Dataset
            Indexed ragged array time series.
        """
        verify_indexed_ragged(ds, self.index_var, self.sample_dim)

        # use instance_id in index variable if instance dimension is a variable
        if self.instance_dim in self.ds:
            self.ds[self.index_var] = self.ds[self.instance_dim][self.ds[
                self.index_var]]

        if self.instance_dim in ds:
            # assume this has been set or not?
            ds = ds.set_coords(self.index_var).set_xindex(self.index_var)
            ds[self.index_var] = (
                self.sample_dim,
                ds[self.instance_dim].values[ds[self.index_var]])

        self._data = xr.combine_nested([self._data, ds], self.sample_dim)

        if self.instance_dim in self.ds:
            instance_ids = self.ds[self.instance_dim].to_numpy()
            lut = np.zeros(instance_ids.max() + 1, dtype=np.int64)
            lut[instance_ids] = np.arange(self.ds[self.instance_dim].size)
            self.ds[self.index_var] = (self.sample_dim,
                                       lut[self.ds[self.index_var]])

        self._set_instance_lut()


def vrange(starts, stops):
    """
    Create concatenated ranges of integers for multiple start/stop values.

    Parameters
    ----------
    starts : numpy.ndarray
        Starts for each range.
    stops : numpy.ndarray
        Stops for each range (same shape as starts).

    Returns
    -------
    ranges : numpy.ndarray
        Concatenated ranges.

    Example
    -------
        >>> starts = [1, 3, 4, 6]
        >>> stops  = [1, 5, 7, 6]
        >>> vrange(starts, stops)
        array([3, 4, 4, 5, 6])
    """
    if starts.shape != stops.shape:
        raise ValueError("starts and stops must have the same shape")

    stops = np.asarray(stops)
    l = stops - starts  # lengths of each range
    return np.repeat(stops - l.cumsum(), l) + np.arange(l.sum())


def pad_to_2d(var: xr.DataArray, x: np.array, y: np.array,
              shape: tuple) -> np.array:
    """
    Pad each time series

    Parameters
    ----------
    var : xarray.DataArray
        1d array to be converted into 2d array.
    x : np.array
        Row indices.
    y : np.array
        Column indices.
    shape : tuple
        Array shape.

    Returns
    -------
    padded : numpy.array
        Padded 2d array.
    """
    padded = np.full(shape, dtype_to_nan[var.dtype], dtype=var.dtype)
    padded[x, y] = var.values
    padded[np.isnan(padded)] = dtype_to_nan[var.dtype]

    return padded


def create_contiguous_ragged():
    """Create CF-compliant contiguous ragged array."""
    count_var = "row_size"
    sample_dim = "obs"
    instance_dim = "loc"

    row_size = np.array([5, 10, 2, 10])
    n_obs = row_size.sum()
    n_instances = row_size.size
    attrs = {"sample_dimension": sample_dim}

    data_vars = {
        "precipitation": ((sample_dim,), 10 * np.random.randn(n_obs)),
        "temperature": ((sample_dim,), 15 + 8 * np.random.randn(n_obs)),
        count_var: ((instance_dim,), row_size, attrs)
    }

    location_id = np.random.choice(
        np.arange(100), size=n_instances, replace=False)
    coords = {
        instance_dim: (instance_dim, np.arange(n_instances)),
        "lon": ((instance_dim,), np.random.randn(n_instances)),
        "lat": ((instance_dim,), np.random.randn(n_instances)),
        "location_id": ((instance_dim,), location_id)
    }

    ds = xr.Dataset(data_vars=data_vars, coords=coords)
    verify_contiguous_ragged(ds, count_var, instance_dim)
    cr = ContiguousRaggedArray(ds, count_var, instance_dim)

    tmp_dir = tempfile.mkdtemp()
    filename = Path(tmp_dir) / "cont_ragged.nc"
    print(filename)
    ds.to_netcdf(filename)


def create_indexed_ragged():
    """Create CF-compliant indexed ragged array."""
    sample_dim = "obs"
    instance_dim = "loc"
    index_var = "locationIndex"

    n_obs = 100
    location_index = np.random.choice(np.arange(10), size=n_obs)
    n_instances = np.unique(location_index).size
    attrs = {"instance_dimension": instance_dim}

    data_vars = {
        "precipitation": ((sample_dim,), 10 * np.random.randn(n_obs)),
        "temperature": ((sample_dim,), 15 + 8 * np.random.randn(n_obs)),
        index_var: ((sample_dim), location_index, attrs),
    }

    location_id = np.random.choice(
        np.arange(100), size=n_instances, replace=False)
    coords = {
        instance_dim: (instance_dim, np.arange(n_instances)),
        "lon": ((instance_dim,), np.random.randn(n_instances)),
        "lat": ((instance_dim,), np.random.randn(n_instances)),
        "location_id": ((instance_dim,), location_id)
    }

    ds = xr.Dataset(data_vars=data_vars, coords=coords)
    verify_indexed_ragged(ds, index_var, sample_dim)
    ir_arr = IndexedRaggedArray(ds, index_var, sample_dim)

    tmp_dir = tempfile.mkdtemp()
    filename = Path(tmp_dir) / "ind_ragged.nc"
    print(filename)
    ds.to_netcdf(filename)


def create_ortho_multi():
    """Create CF-compliant orthomulti array."""
    instance_dim = "loc"
    element_dim = "time"

    n_instances = 10
    n_obs = 10

    data_vars = {
        "precipitation": ((instance_dim, element_dim),
                          10 * np.random.randn(n_instances, n_obs)),
        "temperature": ((instance_dim, element_dim),
                        15 + 8 * np.random.randn(n_instances, n_obs)),
    }

    coords = {
        instance_dim: (instance_dim, np.arange(n_instances)),
        "lon": ((instance_dim,), np.random.randn(n_instances)),
        "lat": ((instance_dim,), np.random.randn(n_instances)),
    }

    ds = xr.Dataset(data_vars=data_vars, coords=coords)
    verify_ortho_multi(ds, instance_dim, element_dim)
    om_arr = OrthoMultiArray(ds, instance_dim, element_dim)

    tmp_dir = tempfile.mkdtemp()
    filename = Path(tmp_dir) / "om.nc"
    print(filename)
    ds.to_netcdf(filename)


def create_point_data():
    """Create CF-compliant point data array."""
    sample_dim = "obs"
    n_obs = 100

    time = np.arange("2024-01-01", "2024-04-10", dtype="datetime64[D]")[:n_obs]
    lat = np.random.uniform(-90, 90, size=n_obs)
    lon = np.random.uniform(-180, 180, size=n_obs)

    data_vars = {
        "precipitation": ((sample_dim,), 10 * np.random.randn(n_obs)),
        "temperature": ((sample_dim,), 15 + 8 * np.random.randn(n_obs)),
    }

    coords = {
        sample_dim: np.arange(n_obs),
        "time": (sample_dim, time),
        "lat": (sample_dim, lat),
        "lon": (sample_dim, lon),
    }

    ds = xr.Dataset(data_vars=data_vars, coords=coords)

    ds["lat"].attrs = {"standard_name": "latitude", "units": "degrees_north"}
    ds["lon"].attrs = {"standard_name": "longitude", "units": "degrees_east"}
    ds["time"].attrs = {"standard_name": "time"}

    verify_point_array(ds, sample_dim)

    tmp_dir = tempfile.mkdtemp()
    filename = Path(tmp_dir) / "point_data.nc"
    print(filename)
    ds.to_netcdf(filename)



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
from pathlib import Path
from datetime import datetime
from tempfile import TemporaryDirectory

import numpy as np
import xarray as xr

from ascat.read_native.ragged_array_ts import CellFileCollection
from ascat.read_native.ragged_array_ts import SwathFileCollection
from ascat.read_native.ragged_array_ts import CellFileCollectionStack

from ascat.read_native.xarray_io import AscatH129Cell
from ascat.read_native.xarray_io import AscatSIG0Cell12500m

from ascat.read_native.xarray_io import AscatH129Swath
from ascat.read_native.xarray_io import AscatSIG0Swath6250m
from ascat.read_native.xarray_io import AscatSIG0Swath12500m

xr.set_options(display_max_rows=100)

TEST_DATA = Path("ascat_test_data")


class TestCellFileCollection(unittest.TestCase):
    """
    Test the merge function
    """

    def assertNpArrayEqual(self, d1, d2):
        try:
            np.testing.assert_equal(d1, d2)
        except AssertionError as e:
            raise self.failureException(e)

    def assertDataSetEqual(self, d1, d2):
        try:
            xr.testing.assert_equal(d1, d2)
        except AssertionError as e:
            raise self.failureException(e)

    # @classmethod
    # def setUpClass(cls):
    #     cls.ctg_old_fname = self.tmpdir / "contiguous_RA_old.nc"
    #     cls.idx_new_fname = self.tmpdir / "indexed_RA_new.nc"
    #     cls.idx_old_fname = self.tmpdir / "indexed_RA_old.nc"
    #     cls.ctg_new_fname = self.tmpdir / "contiguous_RA_new.nc"

    def setUp(self):
        self.temporary_directory = TemporaryDirectory()
        self.tmpdir = Path(self.temporary_directory.name)
        self.cells_sig0_12500_idx = TEST_DATA / "hsaf" / "sig0" / "12500m" / "stack_cells" / "metop_c"

        #TODO make this data
        self.cells_h129_idx = TEST_DATA / "hsaf" / "h129" / "stack_cells"
        self.cells_h129_ctg = TEST_DATA / "hsaf" / "h129" / "merge_cells"

    def tearDown(self):
        self.temporary_directory.cleanup()
        # self.ctg_coll.close()

    def test_from_product_id(self):
        return
        h129 = CellFileCollection.from_product_id(
            self.cells_h129_idx,
            product_id="h129",
        )
        self.assertEqual(h129.ioclass, AscatH129Cell)

        sig0_12_5 = CellFileCollection.from_product_id(
            self.cells_sig0_12500_idx / "20221212000000_20221219000000",
            product_id="sig0_12.5",
        )

        self.assertEqual(sig0_12_5.ioclass, AscatSIG0Cell12500m)

        incorrect = CellFileCollection.from_product_id(
            self.cells_h129_idx,
            product_id="sig0_12.5",
        )

        self.assertEqual(incorrect.ioclass, AscatSIG0Cell12500m)

        # trying to read this data first raises a RuntimeWarning, since it can't find any files
        # for the requested names, then an AttributeError since it tries to read a None object.
        with self.assertRaises(AttributeError):
            with self.assertWarns(RuntimeWarning):
                incorrect.read(
                    start_dt=datetime(2021, 1, 1), end_dt=datetime(2021, 1, 2))

    def test_cells_in_collection(self):
        sig0_12_5 = CellFileCollection.from_product_id(
            self.cells_sig0_12500_idx / "20221212000000_20221219000000",
            product_id="sig0_12.5",
        )

        cells_in_collection = sig0_12_5.cells_in_collection

        expected_cells = [0, 1, 9]
        self.assertEqual(set(cells_in_collection), set(expected_cells))

    def test_read(self):
        coll = CellFileCollection.from_product_id(
            self.cells_sig0_12500_idx / "20221219000000_20230101000000",
            product_id="sig0_12.5",
        )

        cell_5m = coll.read(cell=9)

        # With a new cell size:
        cell_10m = coll.read(cell=9, new_grid=10)

        # the actual read cells should be exactly the same
        # the only difference will be on writing out
        self.assertDataSetEqual(cell_5m, cell_10m)

        id = coll.read(location_id=2148049)
        # assert something

    def test_create_cell_lookup(self):
        coll = CellFileCollection.from_product_id(
            self.cells_sig0_12500_idx / "20221219000000_20230101000000",
            product_id="sig0_12.5")
        self.assertIsNone(coll.cell_lut)
        coll.create_cell_lookup(10)
        self.assertIsNotNone(coll.cell_lut)
        self.assertEqual(set(coll.cell_lut[0]), {0, 1, 36, 37})

    def test_get_cell_path(self):
        source_dir = self.cells_sig0_12500_idx / "20221219000000_20230101000000"
        coll = CellFileCollection.from_product_id(
            source_dir, product_id="sig0_12.5")
        self.assertEqual(coll.get_cell_path(cell=0), source_dir / "0000.nc")
        with self.assertRaises(ValueError):
            # this should raise a value error since cell 2593 is not in the 12.5km FibGrid
            # with a cell size of 5km (max 2592)
            coll.get_cell_path(cell=2593)

        # however, cells within the possible range that simply don't have a file in the
        # folder should work just fine
        self.assertEqual(coll.get_cell_path(cell=100), source_dir / "0100.nc")

    def test_close(self):
        source_dir = self.cells_sig0_12500_idx / "20221219000000_20230101000000"
        coll = CellFileCollection.from_product_id(
            source_dir, product_id="sig0_12.5")
        self.assertIsNone(coll.fid)
        coll.read(cell=0)
        self.assertIsNotNone(coll.fid)
        coll.close()
        self.assertIsNone(coll.fid)


class TestCellFileCollectionStack(unittest.TestCase):

    def assertNanEqual(self, d1, d2):
        try:
            np.testing.assert_equal(d1, d2)
        except AssertionError as e:
            raise self.failureException(e)

    def setUp(self):
        self.temporary_directory = TemporaryDirectory()
        self.tmpdir = Path(self.temporary_directory.name)

        self.cells_sig0_12500_idx_metop_b = TEST_DATA / "hsaf" / "sig0" / "12500m" / "stack_cells" / "metop_b"
        self.cells_sig0_12500_idx_metop_c = TEST_DATA / "hsaf" / "sig0" / "12500m" / "stack_cells" / "metop_c"

        #TODO make this data
        self.cells_h129_idx = TEST_DATA / "hsaf" / "h129" / "stack_cells"
        self.cells_h129_ctg = TEST_DATA / "hsaf" / "h129" / "merge_cells"

    def tearDown(self):
        self.temporary_directory.cleanup()

    def test_from_product_id(self):
        # return
        # h129 = CellFileCollectionStack.from_product_id(self.idx_data_sig0_h129,
        #                                           product_id="h129",)
        # self.assertEqual(h129.ioclass, AscatH129Cell)

        sig0_12_5 = CellFileCollectionStack.from_product_id(
            self.cells_sig0_12500_idx_metop_b,
            product_id="sig0_12.5",
        )

        self.assertEqual(sig0_12_5.ioclass, AscatSIG0Cell12500m)

        product_id_mismatch = CellFileCollectionStack.from_product_id(
            self.cells_sig0_12500_idx_metop_b,
            product_id="h129",
        )

        self.assertEqual(product_id_mismatch.ioclass, AscatH129Cell)

        # TODO currently this does not fail, but it should.
        # with self.assertRaises():
        # incorrect_data = product_id_mismatch.read(cell=0)

    def test_add_collection(self):
        stack = CellFileCollectionStack.from_product_id(
            self.cells_sig0_12500_idx_metop_b, product_id="sig0_12.5")

        self.assertEqual(set(stack.read(cell=1).sat_id.values), {4})

        # merge metop_b with metop_c
        stack.add_collection([self.cells_sig0_12500_idx_metop_c],
                             product_id="sig0_12.5")
        self.assertEqual(set(stack.read(cell=1).sat_id.values), {4, 5})

    def test_read(self):
        stack = CellFileCollectionStack.from_product_id(
            self.cells_sig0_12500_idx_metop_b, product_id="sig0_12.5")

        stack.read(cell=0)
        stack.read(location_id=2148049)
        stack.read(location_id=[1650078, 2148049])

    def test_merge_and_write(self):
        stack = CellFileCollectionStack.from_product_id(
            self.cells_sig0_12500_idx_metop_b, product_id="sig0_12.5")
        stack.merge_and_write(self.tmpdir / "5m", cells=[0, 1, 9])
        stack.merge_and_write(
            self.tmpdir / "10m", cells=[0, 1, 9], out_cell_size=10)

        # test that writing to different grid cell size produces expected output
        written_files_5m = {
            int(f.stem) for f in (self.tmpdir / "5m").glob("*.nc")
        }
        written_files_10m = {
            int(f.stem) for f in (self.tmpdir / "10m").glob("*.nc")
        }
        self.assertEqual(written_files_5m, {0, 1, 9})
        self.assertEqual(written_files_10m, {0, 4})

        # test merge_and_write without multiprocessing
        stack.merge_and_write(self.tmpdir / "5m", cells=[0, 1, 9], processes=1)


class TestSwathFileCollection(unittest.TestCase):

    def assertNanEqual(self, d1, d2):
        try:
            np.testing.assert_equal(d1, d2)
        except AssertionError as e:
            raise self.failureException(e)

    def setUp(self):
        self.temporary_directory = TemporaryDirectory()
        self.tmpdir = Path(self.temporary_directory.name)

        self.idx_data = Path(
            "/home/charriso/test_cells/data-write/RADAR/hsaf/stack_cell_new/metop_a"
        )
        self.ref_idx_data = TEST_DATA / "hsaf" / "sig0/12500m/metop_c" / "stack_cells"
        self.ctg_data = Path(
            "/home/charriso/test_cells/data-write/RADAR/hsaf/stack_cell_merged_new/metop_a"
        )
        # self.swath_data = Path("/home/charriso/test_cells/data-write/RADAR/hsaf/metop_a")
        self.swath_data = TEST_DATA / "hsaf" / "h129" / "swaths"
        self.swath_data_sig0_6_25 = Path("")
        self.swath_data_sig0_12_5 = Path("")

    def tearDown(self):
        self.temporary_directory.cleanup()
        # self.ctg_coll.close()
        # self.idx_coll.close()

    def test_init(self):
        return
        a = SwathFileCollection.from_product_id(
            self.swath_data,
            product_id="h129",
            dask_scheduler="processes",
        )
        start = datetime(2021, 1, 1)
        end = datetime(2021, 1, 2)
        fname = a.get_filenames(start, end)
        outdir = Path(
            "/home/charriso/test_cells/data-write/RADAR/hsaf/tester/metop_a")
        from time import time

        # swath = xr.decode_cf(a.read(start, end, cell=0).load())
        # print(swath["backscatter"].attrs["missing_value"])
        # print(np.any(swath["backscatter"] == swath["backscatter"].attrs["missing_value"], axis=1).sum())
        # print(np.all(np.isin(swath["backscatter"], [swath["backscatter"].attrs["missing_value"], ]), axis=1).sum())
        # print(swath.sel(
        #     obs=~np.any(
        #         swath["backscatter"] == swath["backscatter"].attrs["missing_value"],
        #         axis=1
        #     )
        # ))
        # print(np.all(np.isnan(swath.backscatter), axis=1).sum())
        #########################################################
        st_time = time()
        a.stack(fname, outdir, processes=8)
        print(time() - st_time, "seconds to stack a week of swaths")

    def test_from_product_id(self):
        h129 = SwathFileCollection.from_product_id(
            self.swath_data,
            product_id="h129",
        )
        self.assertEqual(h129.ioclass, AscatH129Swath)

        sig0_6_25 = SwathFileCollection.from_product_id(
            self.swath_data_sig0_6_25,
            product_id="sig0_6.25",
        )
        self.assertEqual(sig0_6_25.ioclass, AscatSIG0Swath6250m)

        sig0_12_5 = SwathFileCollection.from_product_id(
            self.swath_data_sig0_12_5,
            product_id="sig0_12.5",
        )
        self.assertEqual(sig0_12_5.ioclass, AscatSIG0Swath12500m)

        # this works just fine because there are no checks to make sure the product_id actually
        # matches the data. (if those were possible, the product_id would be redundant)
        # However, it will fail at some point when trying to read the data, either because the
        # format for filenames is different and it won't find any files, or because the data
        # has different variables and it will try to read something that doesn't exist.
        incorrect = SwathFileCollection.from_product_id(
            self.swath_data,
            product_id="sig0_12.5",
        )
        self.assertEqual(incorrect.ioclass, AscatSIG0Swath12500m)
        self.assertEqual(
            incorrect.get_filenames(
                datetime(2021, 1, 12), datetime(2021, 1, 13)), [])

        # trying to read this data first raises a RuntimeWarning, since it can't find any files
        # for the requested names, then an AttributeError since it tries to read a None object.
        with self.assertRaises(ValueError):
            with self.assertWarns(RuntimeWarning):
                incorrect.read((datetime(2021, 1, 12), datetime(2021, 1, 13)))

    def test_read_and_process(self):
        swaths = SwathFileCollection.from_product_id(
            self.swath_data,
            product_id="h129",
            dask_scheduler="processes",
        )
        start_p1 = datetime(2021, 1, 12)  #, 0, 0, 0)
        end_p1 = datetime(2021, 1, 13)  #, 0, 3, 0)
        ds = swaths.read((start_p1, end_p1)).load()
        processed_ds = swaths.process(ds)

    def test_continuity(self, run_test=False):
        # testing "stack"
        if not run_test:
            return

        swaths = SwathFileCollection.from_product_id(
            self.swath_data,
            product_id="h129",
            dask_scheduler="processes",
        )
        start_p1 = datetime(2021, 1, 1)
        end_p1 = datetime(2021, 1, 4)
        fnames_p1 = swaths.get_filenames(start_p1, end_p1)
        start_p2 = datetime(2021, 1, 4)
        end_p2 = datetime(2021, 1, 7)
        fnames_p2 = swaths.get_filenames(start_p2, end_p2)
        tester = Path(
            "/home/charriso/test_cells/data-write/RADAR/hsaf/tester/")
        outdir = tester / "cells"
        outdir_p1 = outdir / "p1"
        # outdir_p1.mkdir(exist_ok=True, parents=True)
        outdir_p2 = outdir / "p2"
        # outdir_p2.mkdir(exist_ok=True, parents=True)
        swaths.stack(fnames_p1, outdir_p1, processes=8)
        swaths.stack(fnames_p2, outdir_p2, processes=8)
        #

        cells = CellFileCollectionStack.from_product_id(
            outdir, product_id="h129")
        cell_outdir = tester / "merged_cells"
        cell_outdir.mkdir(exist_ok=True, parents=True)
        # print(cells.read(cell=0))
        cells.merge_and_write(out_dir=cell_outdir)
        max_cell_date = None
        min_cell_date = None
        for file in list(cell_outdir.glob("*.nc")):
            with xr.open_dataset(file, mask_and_scale=True) as written_ds:
                row_sum = written_ds["row_size"].sum()
                num_obs = written_ds["obs"].size
                self.assertEqual(row_sum, num_obs)

                written_ds_max_time = written_ds["time"].values.max()
                written_ds_min_time = written_ds["time"].values.min()
                if max_cell_date is None or written_ds_max_time > max_cell_date:
                    max_cell_date = written_ds_max_time
                if min_cell_date is None or written_ds_min_time < min_cell_date:
                    min_cell_date = written_ds_min_time

        self.assertLessEqual(
            max_cell_date.astype("datetime64[D]"), end_p2.date())
        self.assertGreaterEqual(
            min_cell_date.astype("datetime64[D]"), start_p1.date())

    def test_output(self):
        return
        ref = CellFileCollection.from_product_id(
            self.ref_idx_data / "20221212000000_20221219000000",
            product_id="sig0_12.5")
        outdir = Path(
            "/home/charriso/test_cells/data-write/RADAR/hsaf/tester/metop_a")
        print("ref")
        print(ref.cells_in_collection)
        print(
            ref.read(cell=0,
                     mask_and_scale=False).sel(obs=range(0, 10)).load())

        this = CellFileCollection.from_product_id(outdir, product_id="h129")
        print("this")
        this = this.read(cell=0, mask_and_scale=True).load()
        print(this)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import re
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path

import netCDF4
import xarray as xr
import numpy as np
import dask.array as da

from fibgrid.realization import FibGrid
from ascat.file_handling import ChronFiles

int8_nan = np.iinfo(np.int8).min
uint8_nan = np.iinfo(np.uint8).max
int16_nan = np.iinfo(np.int16).min
uint16_nan = np.iinfo(np.uint16).max
int32_nan = np.iinfo(np.int32).min
uint32_nan = np.iinfo(np.uint32).max
int64_nan = np.iinfo(np.int64).min
uint64_nan = np.iinfo(np.uint64).max
float32_nan = -999999.
float64_nan = -999999.

dtype_to_nan = {
    np.dtype('int8'): int8_nan,
    np.dtype('uint8'): uint8_nan,
    np.dtype('int16'): int16_nan,
    np.dtype('uint16'): uint16_nan,
    np.dtype('int32'): int32_nan,
    np.dtype('uint32'): uint32_nan,
    np.dtype('int64'): int64_nan,
    np.dtype('uint64'): uint64_nan,
    np.dtype('float32'): float32_nan,
    np.dtype('float64'): float64_nan,
    np.dtype('<U1'): None,
    np.dtype('O'): None,
}


def trim_dates(ds, date_range):
    """
    Trim dates of dataset to a given date range. Assumes the time variable is named
    "time", and observation dimension is named "obs"

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset.
    date_range : tuple of datetime.datetime
        Date range to trim to.

    Returns
    -------
    xarray.Dataset
        Dataset with trimmed dates.
    """
    if date_range is None:
        return ds
    start_date = np.datetime64(date_range[0])
    end_date = np.datetime64(date_range[1])
    return ds.isel(obs=(ds.time >= start_date) & (ds.time <= end_date))


def _expand_variable(nc_variable, data, expanding_dim, nc_shape, added_size):
    # Adapted from @hmaarrfk on github: https://github.com/pydata/xarray/issues/1672
    # For time deltas, we must ensure that we use the same encoding as
    # what was previously stored.
    # We likely need to do this as well for variables that had custom
    # encodings too
    data.encoding = dict()
    if hasattr(nc_variable, 'calendar'):
        data.encoding['calendar'] = nc_variable.calendar
    if hasattr(nc_variable, 'calender'):
        data.encoding['calendar'] = nc_variable.calender

    if hasattr(nc_variable, 'dtype'):
        data.encoding['dtype'] = nc_variable.dtype
    if hasattr(nc_variable, 'units'):
        data.encoding['units'] = nc_variable.units
    if hasattr(nc_variable,
               '_FillValue') and data.attrs.get('_FillValue') is None:
        data.encoding['_FillValue'] = nc_variable._FillValue

    data_encoded = xr.conventions.encode_cf_variable(data)

    left_slices = data.dims.index(expanding_dim)
    right_slices = data.ndim - left_slices - 1
    nc_slice = ((slice(None),) * left_slices +
                (slice(nc_shape, nc_shape + added_size),) + (slice(None),) *
                (right_slices))
    nc_variable[nc_slice] = data_encoded.data


def append_to_netcdf(filename, ds_to_append, unlimited_dim):
    """Appends an xarray dataset to an existing netCDF file along a given unlimited dim.

    Parameters
    ----------
    filename : str or Path
        Filename of netCDF file to append to.
    ds_to_append : xarray.Dataset
        Dataset to append.
    unlimited_dim : str or list of str
        Name of the unlimited dimension to append along.

    Raises
    ------
    ValueError
        If more than one unlimited dim is given.
    """
    # By @hmaarrfk on github: https://github.com/pydata/xarray/issues/1672
    with netCDF4.Dataset(filename, mode='a') as nc:
        nc.set_auto_maskandscale(False)
        nc_coord = nc.dimensions[unlimited_dim]
        nc_shape = len(nc_coord)

        added_size = ds_to_append.sizes[unlimited_dim]
        variables, _ = xr.conventions.encode_dataset_coordinates(ds_to_append)

        for name, data in variables.items():
            if unlimited_dim not in data.dims:
                # Nothing to do, data assumed to the identical
                continue

            nc_variable = nc[name]
            _expand_variable(nc_variable, data, unlimited_dim, nc_shape,
                             added_size)


def var_order(ds):
    """
    Returns a reasonable variable order for a ragged array dataset,
    based on that used in existing datasets.

    Puts the count/index variable first depending on the ragged array type,
    then lon, lat, alt, location_id, location_description, and time,
    followed by the rest of the variables in the dataset.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset.

    Returns
    -------
    order : list of str
        List of dataset variable names in the determined order.
    """
    if "row_size" in ds.data_vars:
        first_var = "row_size"
    elif "locationIndex" in ds.data_vars:
        first_var = "locationIndex"
    else:
        raise ValueError("No row_size or locationIndex in ds."
                         " Cannot determine if indexed or ragged")
    order = [
        first_var,
        "lon",
        "lat",
        "alt",
        "location_id",
        "location_description",
        "time",
    ]
    order.extend([v for v in ds.data_vars if v not in order])

    return order


def set_attributes(ds, variable_attributes=None, global_attributes=None):
    """
    Parameters
    ----------
    ds : xarray.Dataset, Path
        Dataset.
    variable_attributes : dict, optional
        User-defined variable attributes to set. Should be a dictionary with format
        `{"varname": {"attr1": "value1", "attr2": "value2"}, "varname2": {"attr1": "value1"}}`
    global_attributes : dict, optional
        User-defined global attributes to set. Should be a dictionary with format
        `{"attr1": "value1", "attr2": "value2"}`

    Returns
    -------
    ds : xarray.Dataset
        Dataset with variable_attributes.
    """
    variable_attributes = variable_attributes or {}

    if "row_size" in ds.data_vars:
        first_var = "row_size"
    elif "locationIndex" in ds.data_vars:
        first_var = "locationIndex"

    first_var_attrs = {
        "row_size": {
            "long_name": "number of observations at this location",
            "sample_dimension": "obs",
        },
        "locationIndex": {
            "long_name": "which location this observation is for",
            "sample_dimension": "locations",
        },
    }

    default_variable_attributes = {
        first_var: first_var_attrs[first_var],
        "lon": {
            "standard_name": "longitude",
            "long_name": "location longitude",
            "units": "degrees_east",
            "valid_range": np.array([-180, 180], dtype=float),
        },
        "lat": {
            "standard_name": "latitude",
            "long_name": "location latitude",
            "units": "degrees_north",
            "valid_range": np.array([-90, 90], dtype=float),
        },
        "alt": {
            "standard_name": "height",
            "long_name": "vertical distance above the surface",
            "units": "m",
            "positive": "up",
            "axis": "Z",
        },
        "time": {
            "standard_name": "time",
            "long_name": "time of measurement",
        },
        "location_id": {},
        "location_description": {},
    }

    variable_attributes = {
        **default_variable_attributes,
        **variable_attributes
    }

    for var, attrs in variable_attributes.items():
        ds[var] = ds[var].assign_attrs(attrs)
        if var in [
                "row_size", "locationIndex", "location_id",
                "location_description"
        ]:
            ds[var].encoding["coordinates"] = None

    global_attributes = global_attributes or {}

    date_created = datetime.now().isoformat(" ", timespec="seconds")
    default_global_attributes = {
        "date_created": date_created,
        "featureType": "timeSeries",
    }
    global_attributes = {**default_global_attributes, **global_attributes}
    for key, item in global_attributes.items():
        ds.attrs[key] = item

    return ds


def create_variable_encodings(ds,
                              custom_variable_encodings=None,
                              custom_dtypes=None):
    """
    Create an encoding dictionary for a dataset, optionally
    overriding the default encoding or adding additional
    encoding parameters.
    New parameters cannot be added to default encoding for
    a variable, only overridden.

    E.g. if you want to add a "units" encoding to "lon",
    you should also pass "dtype", "zlib", "complevel",
    and "_FillValue" if you don't want to lose those.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset.
    custom_variable_encodings : dict, optional
        Custom encodings.

    Returns
    -------
    ds : xarray.Dataset
        Dataset with encodings.
    """
    if custom_variable_encodings is None:
        custom_variable_encodings = {}

    if "row_size" in ds.data_vars:
        # contiguous RA case
        first_var = {"row_size": {"dtype": "int64"}}
        coord_vars = {
            "lon": {"dtype": "float32",},
            "lat": {"dtype": "float32",},
            "alt": {"dtype": "float32",},
        }
    elif "locationIndex" in ds.data_vars:
        # indexed RA case
        first_var = {"locationIndex": {"dtype": "int64"}}
        coord_vars = {
            "lon": {"dtype": "float32"},
            "lat": {"dtype": "float32"},
            "alt": {"dtype": "float32"},
        }
    else:
        # swath file case
        first_var = {}
        coord_vars = {
            "lon": {"dtype": "float32"},
            "lat": {"dtype": "float32"},
            # "longitude": {"dtype": "float32"},
            # "latitude": {"dtype": "float32"},
        }

    # default encodings for coordinates and row_size
    default_encoding = {
        **first_var,
        **coord_vars,
        "location_id": {
            "dtype": "int64",
        },
        "time": {
            "dtype": "float64",
            "units": "days since 1900-01-01 00:00:00",
        },
    }

    for _, var_encoding in default_encoding.items():
        if var_encoding["dtype"] != "int64":
            var_encoding["_FillValue"] = None
        var_encoding["zlib"] = True
        var_encoding["complevel"] = 4

    default_encoding.update({
        var: {
            "dtype": dtype,
            "zlib": bool(np.issubdtype(dtype, np.number)),
            "complevel": 4,
            "_FillValue": dtype_to_nan[dtype],
        } for var, dtype in ds.dtypes.items() if var not in default_encoding
    })

    if custom_dtypes is not None:
        custom_variable_encodings = {
            var: {
                "dtype": custom_dtypes[var],
                "zlib": bool(np.issubdtype(custom_dtypes[var], np.number)),
                "complevel": 4,
                "_FillValue": dtype_to_nan[custom_dtypes[var]],
            } for var in custom_dtypes.names if var in ds.data_vars
        }

    encoding = {**default_encoding, **custom_variable_encodings}

    return encoding


class RaggedXArrayCellIOBase(ABC):
    """Base class for ascat xarray IO classes

    Attributes
    ----------
    source : str, Path, list
        Input filename(s).
    engine : str
        Engine to use for reading/writing files.
    """

    # @classmethod
    # def __init_subclass__(cls):
    #     required_class_attrs = [
    #         "grid_info",
    #         "grid",
    #         "grid_cell_size",
    #         "fn_format",
    #     ]
    #     for var in required_class_attrs:
    #         if not hasattr(cls, var):
    #             raise NotImplementedError(
    #                 f"{cls.__name__} must define required class attribute {var}"
    #             )

    def __init__(self, source, engine, obs_dim="time", **kwargs):
        self.source = source
        self.engine = engine
        self.expected_obs_dim = obs_dim
        if isinstance(source, list):
            # intended for opening multiple files from the /same cell/.
            # Won't work if two files have different locations dimension.
            self._ds = xr.open_mfdataset(
                source,
                engine=engine,
                preprocess=self._preprocess,
                decode_cf=False,
                mask_and_scale=False,
                concat_dim=obs_dim,
                combine="nested",
                data_vars="minimal",
                **kwargs)
        elif isinstance(source, (str, Path)):
            self._ds = self._ensure_obs_dim(
                xr.open_dataset(
                    source,
                    engine=engine,
                    decode_cf=False,
                    mask_and_scale=False,
                    **kwargs))
        elif isinstance(source, xr.Dataset):
            self._ds = self._ensure_obs_dim(source)

        self._kwargs = kwargs

        # if obs_dim in self._ds.dims:
        #     if obs_dim != "obs":
        #         self._ds = self._ds.rename_dims({obs_dim: "obs"})
        # else:
        #     raise ValueError(f"obs_dim '{obs_dim}' not found in dataset")

        # chunks = kwargs.pop("chunks", None)
        # if chunks is not None:
        #     self._ds = self._ds.chunk(chunks)

        if "row_size" in self._ds.data_vars:
            self._ds = self._ensure_indexed(self._ds)

    def _preprocess(self, ds):
        return self._ensure_obs_dim(ds)

    def _ensure_obs_dim(self, ds, preferred_obs_dim="obs"):
        """Rename the observations dimension of a DataSet if necessary."""
        if preferred_obs_dim in ds.dims:
            return ds
        if self.expected_obs_dim in ds.dims:
            if self.expected_obs_dim != preferred_obs_dim:
                ds = ds.rename_dims({self.expected_obs_dim: preferred_obs_dim})
            return ds
        raise ValueError(
            f"obs_dim '{self.expected_obs_dim}' not found in dataset")

    @property
    def date_range(self):
        """
        Return date range of dataset.

        Returns
        -------
        tuple
            Date range of dataset.
        """

        dates = xr.decode_cf(self._ds[["time"]], mask_and_scale=False)
        return dates.time.min().values, dates.time.max().values

    @abstractmethod
    def read(self, location_id=None, **kwargs):
        """
        Read data from file. Should be implemented by subclasses.

        Parameters
        ----------
        location_id : int, list, optional
            Location id(s) to read.
        **kwargs
            Additional keyword arguments passed to the read method.

        Returns
        -------
        xarray.Dataset
            Dataset containing the data for any specified location_id(s),
            or all location_ids in the file if none are specified.
        """
        raise NotImplementedError

    @abstractmethod
    def write(self, filename, ra_type, **kwargs):
        """
        Write data to file. Should be implemented by subclasses.

        Parameters
        ----------
        filename : str, Path
            Filename to write data to.
        ra_type : str, optional
            Type of ragged array to write.
        **kwargs
            Additional keyword arguments passed to the write method.
        """
        raise NotImplementedError

    def _ensure_indexed(self, ds):
        """
        Convert a contiguous dataset to indexed dataset,
        if necessary. Indexed datasets pass through.

        Ragged array type is determined by the presence of
        either a "row_size" or "locationIndex" variable
        (for contiguous and indexed arrays, respectively).

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset in indexed or contiguous ragged array format.

        Returns
        -------
        xarray.Dataset
            Dataset in indexed ragged array format.
        """
        if ds is None or "locationIndex" in ds.data_vars:
            return ds

        row_size = np.where(ds["row_size"].values > 0, ds["row_size"].values,
                            0)

        locationIndex = np.repeat(np.arange(row_size.size), row_size)
        ds["locationIndex"] = ("obs", locationIndex)
        ds = ds.drop_vars(["row_size"])

        return ds[self._var_order(ds)]

    def _ensure_contiguous(self, ds):
        """
        Convert an indexed dataset to contiguous dataset,
        if necessary. Contiguous datasets pass through.

        Ragged array type is determined by the presence of
        either a "row_size" or "locationIndex" variable
        (for contiguous and indexed arrays, respectively).

        Parameters
        ----------
        ds : xarray.Dataset, Path
            Dataset in indexed or contiguous ragged array format.

        Returns
        -------
        xarray.Dataset
            Dataset in contiguous ragged array format.
        """
        if ds is None or "row_size" in ds.data_vars:
            return ds

        ds = ds.sortby(["locationIndex", "time"])
        idxs, sizes = np.unique(ds.locationIndex, return_counts=True)
        row_size = np.zeros_like(ds.location_id.values)
        row_size[idxs] = sizes
        ds["row_size"] = ("locations", row_size)
        ds = ds.drop_vars(["locationIndex"])
        return ds[self._var_order(ds)]

    def close(self):
        """Close file."""
        if self._ds is not None:
            self._ds.close()
            self._ds = None

    def __enter__(self):
        """Context manager initialization."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the runtime context related to this object."""
        self.close()


class AscatNetCDFCellBase(RaggedXArrayCellIOBase):

    def __init__(self, filename, **kwargs):
        super().__init__(filename, "netcdf4", **kwargs)
        self.custom_variable_attrs = None
        self.custom_global_attrs = None
        self.custom_variable_encodings = None

    def read(self, date_range=None, location_id=None, mask_and_scale=True):
        """Read data from netCDF4 file.

        Read all or a subset of data from a netCDF4 file, with subset specified by the
        `location_id` argument.

        Parameters
        ----------
        date_range : tuple of datetime.datetime, optional
            Date range to read data for. If None, all data is read.
        location_id : int or list of int.
            The location_id(s) to read data for. If None, all data is read.
            Default is None.
        mask_and_scale : bool, optional
            If True, mask and scale the data according to its `scale_factor` and
            `_FillValue`/`missing_value` before returning. Default: True.
        """
        if location_id is not None:
            if isinstance(location_id, (int, np.int64, np.int32)):
                ds = self._read_location_id(location_id)
            elif isinstance(location_id, (list, np.ndarray)):
                ds = self._read_location_ids(location_id)
            else:
                raise ValueError("location_id must be int or list of ints")
        else:
            ds = self._ds

        return trim_dates(
            xr.decode_cf(ds, mask_and_scale=mask_and_scale), date_range)

    def write(self, filename, ra_type="indexed", **kwargs):
        """
        Write data to a netCDF file.

        Parameters
        ----------
        filename : str
            Output filename.
        ra_type : str, optional
            Type of ragged array to write. Default is "contiguous".
        **kwargs : dict
            Additional keyword arguments passed to xarray.to_netcdf().
        """

        if ra_type not in ["contiguous", "indexed"]:
            raise ValueError("ra_type must be 'contiguous' or 'indexed'")
        out_ds = self._ds
        if ra_type == "contiguous":
            out_ds = self._ensure_contiguous(out_ds)

        out_ds = out_ds[self._var_order(out_ds)]

        custom_variable_attrs = self._kwargs.get(
            "attributes", None) or self.custom_variable_attrs
        custom_global_attrs = self._kwargs.get(
            "global_attributes", None) or self.custom_global_attrs
        out_ds = self._set_attributes(out_ds, custom_variable_attrs,
                                      custom_global_attrs)

        custom_variable_encodings = self._kwargs.get(
            "encoding", None) or self.custom_variable_encodings
        encoding = self._create_variable_encodings(out_ds,
                                                   custom_variable_encodings)
        out_ds.encoding["unlimited_dims"] = ["obs"]

        for var, var_encoding in encoding.items():
            if "_FillValue" in var_encoding and "_FillValue" in out_ds[
                    var].attrs:
                del out_ds[var].attrs["_FillValue"]

        out_ds.to_netcdf(filename, encoding=encoding, **kwargs)

    def _read_location_id(self, location_id):
        """
        Read data for a single location_id
        """
        if location_id not in self._ds.location_id.values:
            return None
        ds = self._ds
        idx = np.where(ds.location_id.values == location_id)[0][0]
        locationIndex = np.where(ds.locationIndex.values == idx)[0]
        ds = ds.isel(obs=locationIndex, locations=idx)

        return ds

    def _read_location_ids(self, location_ids):
        """
        Read data for a list of location_ids.
        If there are location_ids not in the dataset,
        they are ignored.
        """
        if not np.any(np.isin(location_ids, self._ds.location_id.values)):
            return None
        ds = self._ds
        idxs = np.where(np.isin(ds.location_id.values, location_ids))[0]
        locationIndex = np.where(np.isin(ds.locationIndex.values, idxs))[0]
        ds = ds.isel(obs=locationIndex, locations=idxs)

        return ds

    @staticmethod
    def _var_order(ds):
        """
        A wrapper for var_order, which can be overridden in a child class
        if different logic for this function is needed.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset.

        Returns
        -------
        list of str
            List of dataset variable names in the determined order (result of var_order)
        """

        return var_order(ds)

    @staticmethod
    def _set_attributes(ds, variable_attributes=None, global_attributes=None):
        """
        A wrapper for xarray_io.set_attributes, which can be overriden in a child class
        if different logic for this function is needed.
        Parameters
        ----------
        ds : xarray.Dataset, Path
            Dataset.
       variable_attributes : dict, optional
           variable_attributes.

        Returns
        -------
        ds : xarray.Dataset
            Dataset withvariable_attributes.
        """
        return set_attributes(ds, variable_attributes, global_attributes)

    @staticmethod
    def _create_variable_encodings(ds, custom_variable_encodings=None, custom_dtypes=None):
        """
        A wrapper for xarray_io.create_variable_encodings. This can be overridden in a child class
        if different logic for this function is needed.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset.
        custom_variable_encodings : dict, optional
            Custom encodings.

        Returns
        -------
        ds : xarray.Dataset
            Dataset with encodings.
        """
        return create_variable_encodings(ds, custom_variable_encodings, custom_dtypes)


class SwathIOBase(ABC):
    """
    Base class for reading swath data.
    Writes ragged array cell data in indexed or contiguous format.
    """

    beams_vars = []

    @classmethod
    def __init_subclass__(cls):
        required_class_attrs = [
            "fn_pattern",
            "sf_pattern",
            "date_format",
            "grid",
            "grid_cell_size",
            "cell_fn_format",
            "ts_dtype",
        ]
        for var in required_class_attrs:
            if not hasattr(cls, var):
                raise NotImplementedError(
                    f"{cls.__name__} must define required class attribute {var}"
                )

    @classmethod
    def chron_files(cls, path):
        """
        Return a ChronFiles object for this class type based on a path.
        """
        return ChronFiles(
            path,
            cls,
            cls.fn_pattern,
            cls.sf_pattern,
            None,
            True,
            cls.fn_read_fmt,
            cls.sf_read_fmt,
        )

    @staticmethod
    @abstractmethod
    def fn_read_fmt():
        """
        TODO: figure out a sane way to describe what this does.
        Also decide if this /needs/ to be enforced. If the user
        doesn't want to use all the filesearch functionality (or
        if they want to use their own filesearch logic), then
        they should still be able to use this class. They could
        of course override this and just return None, but that
        seems like a hack.
        """
        pass

    @staticmethod
    @abstractmethod
    def sf_read_fmt():
        """
        TODO: same as above
        """
        pass

    def __init__(self, source, engine, **kwargs):
        self.source = source
        self.engine = engine
        self._cell_vals = None

        # if filename is an iterable, use open_mfdataset
        # else use open_dataset
        chunks = kwargs.pop("chunks", None)
        if isinstance(source, list):
            self._ds = xr.open_mfdataset(
                source,
                engine=engine,
                preprocess=self._preprocess,
                decode_cf=False,
                concat_dim="obs",
                combine="nested",
                chunks=(chunks or None),
                combine_attrs=self.combine_attributes,
                mask_and_scale=False,
                **kwargs)
            if chunks is not None:
                self._isin = da.isin
            else:
                self._isin = np.in1d

            chunks = None

        elif isinstance(source, (str, Path)):
            self._ds = xr.open_dataset(
                source,
                engine=engine,
                **kwargs,
                decode_cf=False,
                mask_and_scale=False)
            self._ds = self._preprocess(self._ds)

        elif isinstance(source, xr.Dataset):
            self._ds = source
            self._ds = self._preprocess(self._ds)

        self._kwargs = kwargs

        if chunks is not None:
            self._ds = self._ds.chunk(chunks)

        # the swath files have a typo in the calendar attribute
        # this fixes that and lets xarray pick it up when decoding the time variable
        # I'm not sure if it really makes a difference but this is more correct.
        if "calender" in self._ds.time.attrs:
            self._ds.time.attrs["calendar"] = self._ds.time.attrs.pop(
                "calender")

    def read(self,
             cell=None,
             location_id=None,
             mask_and_scale=True,
             lookup_vector=None):
        """
        Returns data for a cell or location_id if specified, or for the entire
        swath file if not specified.

        Parameters
        ----------
        cell : int, optional
            Cell to read data for.
        location_id : int, optional
            Location id to read data for.
        mask_and_scale : bool, optional
            Whether to mask and scale the data. Default is True.
        """
        if location_id is not None:
            out_data = self._read_location_ids(
                location_id, lookup_vector=lookup_vector)
        elif cell is not None:
            out_data = self._read_cell(cell)
        else:
            out_data = self._ds
        return xr.decode_cf(out_data, mask_and_scale=mask_and_scale)

    def write(self, filename, mode="w", **kwargs):
        out_ds = self._ds
        out_ds = out_ds[self._var_order(out_ds)]
        out_ds = self._set_attributes(out_ds)
        # out_encoding = self._create_variable_encodings(out_ds, custom_dtypes = self.ts_dtype)

        if mode == "a":
            if os.path.exists(filename):
                append_to_netcdf(filename, out_ds, unlimited_dim="obs")
                return

        out_encoding = self._create_variable_encodings(
            out_ds, custom_dtypes=self.ts_dtype)
        out_ds.to_netcdf(
            filename, unlimited_dims=["obs"], encoding=out_encoding, **kwargs)

    def _read_location_ids(self, location_ids, lookup_vector=None):
        """
        Read data for a list of location_ids.

        Parameters
        ----------
        location_ids : list or 1D array of int
            Location ids to read data for.
        lookup_vector : 1D array of int, optional
            Lookup vector for location_ids. Should make the search much faster by eliminating
            the need for np.isin. Default is None.
        """
        if lookup_vector is not None:
            return self._ds.sel(obs=lookup_vector[self._ds.location_id.values])

        idxs = self._isin(self._ds.location_id, location_ids)
        idxs = np.array([
            1 if id in location_ids else 0
            for id in self._ds.location_id.values
        ])
        if not np.any(idxs):
            return None
        ds = self._ds.isel(obs=idxs)
        return ds

    def _read_cell(self, cell):
        """
        Read data for a single cell
        """

        if self._cell_vals is None:
            self._cell_vals = self.grid.gpi2cell(self._ds.location_id.values)

        if isinstance(cell, (list, np.ndarray)):
            idxs = np.in1d(self._cell_vals, cell)
        elif isinstance(cell, (int, np.int64, np.int32)):
            idxs = np.where(self._cell_vals == cell)[0]
        else:
            raise ValueError("cell must be int or list of ints")
        if not np.any(idxs):
            return None

        ds = self._ds.isel(obs=idxs)
        return ds

    @staticmethod
    def _var_order(ds):
        """
        A wrapper for var_order, which can be overridden in a child class
        if different logic for this function is needed.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset.

        Returns
        -------
        list of str
            List of dataset variable names in the determined order (result of var_order)
        """

        return var_order(ds)

    @staticmethod
    def _set_attributes(ds, variable_attributes=None, global_attributes=None):
        """
        A wrapper for xarray_io.set_attributes, which can be overriden in a child class
        if different logic for this function is needed.
        Parameters
        ----------
        ds : xarray.Dataset, Path
            Dataset.
        variable_attributes : dict, optional
           variable_attributes.

        Returns
        -------
        ds : xarray.Dataset
            Dataset with variable_attributes and global_attributes.
        """
        return set_attributes(ds, variable_attributes, global_attributes)

    @staticmethod
    def _create_variable_encodings(ds,
                                   custom_variable_encodings=None,
                                   custom_dtypes=None):
        """
        A wrapper for xarray_io.create_variable_encodings. This can be overridden in a child class
        if different logic for this function is needed.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset.
        custom_variable_encodings : dict, optional
            Custom encodings.

        Returns
        -------
        ds : xarray.Dataset
            Dataset with encodings.
        """
        return create_variable_encodings(ds, custom_variable_encodings, custom_dtypes)

    @staticmethod
    def _preprocess(ds):
        """
        To use custom logic for combining attributes in xarray, you need to write a function
        that takes a list of attribute dictionaries as an argument.
        """
        ds.attrs["global_attributes_flag"] = 1
        if "spacecraft" in ds.attrs:
            # Assumption: the spacecraft attribute is something like "metop-a"
            sat_id = {"a": 3, "b": 4, "c": 5}
            sat = ds.attrs["spacecraft"][-1].lower()
            ds["sat_id"] = ("obs",
                            np.repeat(sat_id[sat], ds["location_id"].size))
            del ds.attrs["spacecraft"]
        return ds

    @staticmethod
    def combine_attributes(attrs_list, context):
        """
        Decides which attributes to keep when merging swath files.

        Parameters
        ----------
        attrs_list : list of dict
            List of attributes dictionaries.
        context : None
            This currently is None, but will eventually be passed information about
            the context in which this was called.
            (see https://github.com/pydata/xarray/issues/6679#issuecomment-1150946521)

        Returns
        -------
        """
        # we don't need to pass on anything from global attributes
        if "global_attributes_flag" in attrs_list[0].keys():
            return None

        else:
            variable_attrs = attrs_list
            # this code taken straight from xarray/core/merge.py
            # Replicates the functionality of "drop_conflicts"
            # but just for variable attributes
            result = {}
            dropped_keys = set()
            for attrs in variable_attrs:
                result.update({
                    key: value
                    for key, value in attrs.items()
                    if key not in result and key not in dropped_keys
                })
                result = {
                    key: value
                    for key, value in result.items()
                    if key not in attrs or
                    xr.core.utils.equivalent(attrs[key], value)
                }
                dropped_keys |= {key for key in attrs if key not in result}
            return result

    def contains_location_ids(self, location_ids=None, lookup_vector=None):
        """
        Check if the dataset contains any of the given location_ids.

        Parameters
        ----------
        location_ids : list of int
            Location ids to check.

        Returns
        -------
        bool
            True if the dataset contains any of the given location_ids, False otherwise.
        """
        if lookup_vector is not None:
            return lookup_vector[self._ds.location_id.values].any()

        if location_ids is not None:
            return np.any(
                np.in1d(
                    np.unique(self._ds.location_id.values),
                    location_ids,
                    assume_unique=True))
        else:
            raise ValueError(
                "Must provide either location_ids or lookup_vector")

    def __enter__(self):
        """
        Context manager initialization.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit the runtime context related to this object.
        """
        self.close()

    def close(self):
        """
        Close the dataset.
        """
        self._ds.close()


class CellGridCache:
    """
    Cache for CellGrid objects.
    """

    def __init__(self):
        self.grids = {}

    def fetch_or_store(self, key, cell_grid_type=None, *args):
        """
        Fetch a CellGrid object from the cache given a key,
        or store a new one.

        Parameters
        ----------
        """
        if key not in self.grids:
            if cell_grid_type is None:
                raise ValueError(
                    "Key not in cache, please specify cell_grid_type and arguments"
                    " to create a new CellGrid object and add it to the cache under"
                    " the given key.")
            self.grids[key] = dict()
            self.grids[key]["grid"] = cell_grid_type(*args)
            self.grids[key]["possible_cells"] = self.grids[key][
                "grid"].get_cells()
            self.grids[key]["max_cell"] = self.grids[key][
                "possible_cells"].max()
            self.grids[key]["min_cell"] = self.grids[key][
                "possible_cells"].min()

        return self.grids[key]


grid_cache = CellGridCache()


# Define dataset-specific classes.
class AscatH129Cell(AscatNetCDFCellBase):
    grid_info = grid_cache.fetch_or_store("Fib6.25", FibGrid, 6.25)
    grid = grid_info["grid"]
    grid_cell_size = 5
    fn_format = "{:04d}.nc"
    possible_cells = grid_info["possible_cells"]
    max_cell = grid_info["max_cell"]
    min_cell = grid_info["min_cell"]

    def __init__(self, filename, **kwargs):
        super().__init__(filename, obs_dim="obs", **kwargs)
        self.custom_variable_attrs = None
        self.custom_global_attrs = None
        self.custom_variable_encodings = None


class AscatH129v1Cell(AscatNetCDFCellBase):
    grid_info = grid_cache.fetch_or_store("Fib6.25", FibGrid, 6.25)
    grid = grid_info["grid"]
    grid_cell_size = 5
    fn_format = "{:04d}.nc"
    possible_cells = grid_info["possible_cells"]
    max_cell = grid_info["max_cell"]
    min_cell = grid_info["min_cell"]

    def __init__(self, filename, **kwargs):
        super().__init__(filename, obs_dim="obs", **kwargs)
        self.custom_variable_attrs = None
        self.custom_global_attrs = None
        self.custom_variable_encodings = None


class AscatH121v1Cell(AscatNetCDFCellBase):
    grid_info = grid_cache.fetch_or_store("Fib12.5", FibGrid, 12.5)
    grid = grid_info["grid"]
    grid_cell_size = 5
    fn_format = "{:04d}.nc"
    possible_cells = grid_info["possible_cells"]
    max_cell = grid_info["max_cell"]
    min_cell = grid_info["min_cell"]

    def __init__(self, filename, **kwargs):
        super().__init__(filename, obs_dim="obs", **kwargs)
        self.custom_variable_attrs = None
        self.custom_global_attrs = None
        self.custom_variable_encodings = None


class AscatH122Cell(AscatNetCDFCellBase):
    grid_info = grid_cache.fetch_or_store("Fib6.25", FibGrid, 6.25)
    grid = grid_info["grid"]
    grid_cell_size = 5
    fn_format = "{:04d}.nc"
    possible_cells = grid_info["possible_cells"]
    max_cell = grid_info["max_cell"]
    min_cell = grid_info["min_cell"]

    def __init__(self, filename, **kwargs):
        super().__init__(filename, obs_dim="obs", **kwargs)
        self.custom_variable_attrs = None
        self.custom_global_attrs = None
        self.custom_variable_encodings = None


class AscatSIG0Cell6250m(AscatNetCDFCellBase):
    grid_info = grid_cache.fetch_or_store("Fib6.25", FibGrid, 6.25)
    grid = grid_info["grid"]
    grid_cell_size = 5
    fn_format = "{:04d}.nc"
    possible_cells = grid_info["possible_cells"]
    max_cell = grid_info["max_cell"]
    min_cell = grid_info["min_cell"]

    def __init__(self, filename, **kwargs):
        super().__init__(filename, obs_dim="obs", **kwargs)
        self.custom_variable_attrs = None
        self.custom_global_attrs = None
        self.custom_variable_encodings = None


class AscatSIG0Cell12500m(AscatNetCDFCellBase):
    grid_info = grid_cache.fetch_or_store("Fib12.5", FibGrid, 12.5)
    grid = grid_info["grid"]
    grid_cell_size = 5
    fn_format = "{:04d}.nc"
    possible_cells = grid_info["possible_cells"]
    max_cell = grid_info["max_cell"]
    min_cell = grid_info["min_cell"]

    def __init__(self, filename, **kwargs):
        super().__init__(filename, obs_dim="obs", **kwargs)
        self.custom_variable_attrs = None
        self.custom_global_attrs = None
        self.custom_variable_encodings = None


class AscatH129Swath(SwathIOBase):
    fn_pattern = "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-6.25-H129_C_LIIB_{date}_{placeholder}_{placeholder1}____.nc"
    sf_pattern = {"satellite_folder": "metop_[abc]", "year_folder": "{year}"}
    date_format = "%Y%m%d%H%M%S"
    grid_sampling_km = 6.25
    grid = grid_cache.fetch_or_store("Fib6.25", FibGrid,
                                     grid_sampling_km)["grid"]
    grid_cell_size = 5
    cell_fn_format = "{:04d}.nc"
    beams_vars = ["backscatter", "incidence_angle", "azimuth_angle", "kp"]
    ts_dtype = np.dtype([
        ("sat_id", np.int8),
        ("as_des_pass", np.int8),
        ("swath_indicator", np.int8),
        ("backscatter_for", np.float32),
        ("backscatter_mid", np.float32),
        ("backscatter_aft", np.float32),
        ("incidence_angle_for", np.float32),
        ("incidence_angle_mid", np.float32),
        ("incidence_angle_aft", np.float32),
        ("azimuth_angle_for", np.float32),
        ("azimuth_angle_mid", np.float32),
        ("azimuth_angle_aft", np.float32),
        ("kp_for", np.float32),
        ("kp_mid", np.float32),
        ("kp_aft", np.float32),
        ("surface_soil_moisture", np.float32),
        ("surface_soil_moisture_noise", np.float32),
        ("backscatter40", np.float32),
        ("slope40", np.float32),
        ("curvature40", np.float32),
        ("surface_soil_moisture_sensitivity", np.float32),
        ("correction_flag", np.uint8),
        ("processing_flag", np.uint8),
        ("surface_flag", np.uint8),
        ("snow_cover_probability", np.int8),
        ("frozen_soil_probability", np.int8),
        ("wetland_fraction", np.int8),
        ("topographic_complexity", np.int8),
    ])

    @staticmethod
    def fn_read_fmt(timestamp):
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": "[ABC]",
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp):
        return {
            "satellite_folder": {
                "satellite": "metop_[abc]"
            },
            "year_folder": {
                "year": f"{timestamp.year}"
            },
        }

    def __init__(self, filename, **kwargs):
        super().__init__(filename, "netcdf4", **kwargs)


class AscatH129v1Swath(SwathIOBase):
    fn_pattern = "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-6.25km-H129_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc"
    sf_pattern = {"satellite_folder": "metop_[abc]", "year_folder": "{year}"}
    date_format = "%Y%m%d%H%M%S"
    grid_sampling_km = 6.25
    grid = grid_cache.fetch_or_store("Fib6.25", FibGrid,
                                     grid_sampling_km)["grid"]
    grid_cell_size = 5
    cell_fn_format = "{:04d}.nc"
    beams_vars = []
    ts_dtype = np.dtype([
        ("sat_id", np.int8),
        ("as_des_pass", np.int8),
        ("swath_indicator", np.int8),
        ("surface_soil_moisture", np.float32),
        ("surface_soil_moisture_noise", np.float32),
        ("backscatter40", np.float32),
        ("slope40", np.float32),
        ("curvature40", np.float32),
        ("surface_soil_moisture_sensitivity", np.float32),
        ("backscatter_flag", np.uint8),
        ("correction_flag", np.uint8),
        ("processing_flag", np.uint8),
        ("surface_flag", np.uint8),
        ("snow_cover_probability", np.int8),
        ("frozen_soil_probability", np.int8),
        ("wetland_fraction", np.int8),
        ("topographic_complexity", np.int8),
        ("subsurface_scattering_probability", np.int8),
    ])

    @staticmethod
    def fn_read_fmt(timestamp):
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": "[ABC]",
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp):
        return {
            "satellite_folder": {
                "satellite": "metop_[abc]"
            },
            "year_folder": {
                "year": f"{timestamp.year}"
            },
        }

    def __init__(self, filename, **kwargs):
        super().__init__(filename, "netcdf4", **kwargs)


class AscatH121v1Swath(SwathIOBase):
    fn_pattern = "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-12.5km-H121_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc"
    sf_pattern = {"satellite_folder": "metop_[abc]", "year_folder": "{year}"}
    date_format = "%Y%m%d%H%M%S"
    grid_sampling_km = 12.5
    grid = grid_cache.fetch_or_store("Fib12.5", FibGrid,
                                     grid_sampling_km)["grid"]
    grid_cell_size = 5
    cell_fn_format = "{:04d}.nc"
    beams_vars = []
    ts_dtype = np.dtype([
        ("sat_id", np.int8),
        ("as_des_pass", np.int8),
        ("swath_indicator", np.int8),
        ("surface_soil_moisture", np.float32),
        ("surface_soil_moisture_noise", np.float32),
        ("backscatter40", np.float32),
        ("slope40", np.float32),
        ("curvature40", np.float32),
        ("surface_soil_moisture_sensitivity", np.float32),
        ("backscatter_flag", np.uint8),
        ("correction_flag", np.uint8),
        ("processing_flag", np.uint8),
        ("surface_flag", np.uint8),
        ("snow_cover_probability", np.int8),
        ("frozen_soil_probability", np.int8),
        ("wetland_fraction", np.int8),
        ("topographic_complexity", np.int8),
        ("subsurface_scattering_probability", np.int8),
    ])

    @staticmethod
    def fn_read_fmt(timestamp):
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": "[ABC]",
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp):
        return {
            "satellite_folder": {
                "satellite": "metop_[abc]"
            },
            "year_folder": {
                "year": f"{timestamp.year}"
            },
        }

    def __init__(self, filename, **kwargs):
        super().__init__(filename, "netcdf4", **kwargs)


class AscatH122Swath(SwathIOBase):
    fn_pattern = "ascat_ssm_nrt_6.25km_{placeholder}Z_{date}Z_metop-{sat}_h122.nc"
    sf_pattern = {"satellite_folder": "metop_[abc]", "year_folder": "{year}"}
    date_format = "%Y%m%d%H%M%S"
    grid_sampling_km = 6.25
    grid = grid_cache.fetch_or_store("Fib6.25", FibGrid,
                                     grid_sampling_km)["grid"]
    grid_cell_size = 5
    cell_fn_format = "{:04d}.nc"
    beams_vars = []
    ts_dtype = np.dtype([
        ("sat_id", np.int64),
        ("as_des_pass", np.int8),
        ("swath_indicator", np.int8),
        ("surface_soil_moisture", np.float32),
        ("surface_soil_moisture_noise", np.float32),
        ("sigma40", np.float32),
        ("sigma40_noise", np.float32),
        ("slope40", np.float32),
        ("slope40_noise", np.float32),
        ("curvature40", np.float32),
        ("curvature40_noise", np.float32),
        ("dry40", np.float32),
        ("dry40_noise", np.float32),
        ("wet40", np.float32),
        ("wet40_noise", np.float32),
        ("surface_soil_moisture_sensitivity", np.float32),
        ("surface_soil_moisture_climatology", np.float32),
        ("correction_flag", np.uint8),
        ("processing_flag", np.uint8),
        ("snow_cover_probability", np.int8),
        ("frozen_soil_probability", np.int8),
        ("wetland_fraction", np.int8),
        ("topographic_complexity", np.int8),
    ])

    @staticmethod
    def fn_read_fmt(timestamp):
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": "[ABC]",
            "placeholder": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp):
        return {
            "satellite_folder": {
                "satellite": "metop_[abc]"
            },
            "year_folder": {
                "year": f"{timestamp.year}"
            },
        }

    def __init__(self, filename, **kwargs):
        super().__init__(filename, "netcdf4", **kwargs)


class AscatSIG0Swath6250m(SwathIOBase):
    """
    Class for reading ASCAT sigma0 swath data and writing it to cells.
    """
    fn_pattern = "W_IT-HSAF-ROME,SAT,SIG0-ASCAT-METOP{sat}-6.25_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc"
    sf_pattern = {"satellite_folder": "metop_[abc]", "year_folder": "{year}"}
    date_format = "%Y%m%d%H%M%S"
    grid_sampling_km = 6.25
    grid = grid_cache.fetch_or_store("Fib6.25", FibGrid,
                                     grid_sampling_km)["grid"]
    grid_cell_size = 5
    cell_fn_format = "{:04d}.nc"
    beams_vars = [
        "backscatter",
        "backscatter_std",
        "incidence_angle",
        "azimuth_angle",
        "kp",
        "n_echos",
        "all_backscatter",
        "all_backscatter_std",
        "all_incidence_angle",
        "all_azimuth_angle",
        "all_kp",
        "all_n_echos",
    ]
    ts_dtype = np.dtype([
        ("sat_id", np.int8),
        ("as_des_pass", np.int8),
        ("swath_indicator", np.int8),
        ("backscatter_for", np.float32),
        ("backscatter_mid", np.float32),
        ("backscatter_aft", np.float32),
        ("backscatter_std_for", np.float32),
        ("backscatter_std_mid", np.float32),
        ("backscatter_std_aft", np.float32),
        ("incidence_angle_for", np.float32),
        ("incidence_angle_mid", np.float32),
        ("incidence_angle_aft", np.float32),
        ("azimuth_angle_for", np.float32),
        ("azimuth_angle_mid", np.float32),
        ("azimuth_angle_aft", np.float32),
        ("kp_for", np.float32),
        ("kp_mid", np.float32),
        ("kp_aft", np.float32),
        ("n_echos_for", np.int8),
        ("n_echos_mid", np.int8),
        ("n_echos_aft", np.int8),
        ("all_backscatter_for", np.float32),
        ("all_backscatter_mid", np.float32),
        ("all_backscatter_aft", np.float32),
        ("all_backscatter_std_for", np.float32),
        ("all_backscatter_std_mid", np.float32),
        ("all_backscatter_std_aft", np.float32),
        ("all_incidence_angle_for", np.float32),
        ("all_incidence_angle_mid", np.float32),
        ("all_incidence_angle_aft", np.float32),
        ("all_azimuth_angle_for", np.float32),
        ("all_azimuth_angle_mid", np.float32),
        ("all_azimuth_angle_aft", np.float32),
        ("all_kp_for", np.float32),
        ("all_kp_mid", np.float32),
        ("all_kp_aft", np.float32),
        ("all_n_echos_for", np.int8),
        ("all_n_echos_mid", np.int8),
        ("all_n_echos_aft", np.int8),
    ])

    @staticmethod
    def fn_read_fmt(timestamp):
        """
        Format a timestamp to search as YYYYMMDD*, for use in a regex
        that will match all files covering a single given date.

        Parameters
        ----------
        timestamp: datetime.datetime
            Timestamp to format

        Returns
        -------
        dict
            Dictionary of formatted strings
        """
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": "[ABC]",
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp):
        return {
            "satellite_folder": {
                "satellite": "metop_[abc]"
            },
            "year_folder": {
                "year": f"{timestamp.year}"
            },
        }

    def __init__(self, filename, **kwargs):
        super().__init__(filename, "netcdf4", **kwargs)


class AscatSIG0Swath12500m(SwathIOBase):
    """
    Class for reading and writing ASCAT sigma0 swath data.
    """
    fn_pattern = "W_IT-HSAF-ROME,SAT,SIG0-ASCAT-METOP{sat}-12.5_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc"
    sf_pattern = {"satellite_folder": "metop_[abc]", "year_folder": "{year}"}
    date_format = "%Y%m%d%H%M%S"
    grid_sampling_km = 12.5
    grid = grid_cache.fetch_or_store("Fib12.5", FibGrid,
                                     grid_sampling_km)["grid"]
    grid_cell_size = 5
    cell_fn_format = "{:04d}.nc"
    beams_vars = [
        "backscatter",
        "backscatter_std",
        "incidence_angle",
        "azimuth_angle",
        "kp",
        "n_echos",
        "all_backscatter",
        "all_backscatter_std",
        "all_incidence_angle",
        "all_azimuth_angle",
        "all_kp",
        "all_n_echos",
    ]
    ts_dtype = np.dtype([
        ("sat_id", np.int8),
        ("as_des_pass", np.int8),
        ("swath_indicator", np.int8),
        ("backscatter_for", np.float32),
        ("backscatter_mid", np.float32),
        ("backscatter_aft", np.float32),
        ("backscatter_std_for", np.float32),
        ("backscatter_std_mid", np.float32),
        ("backscatter_std_aft", np.float32),
        ("incidence_angle_for", np.float32),
        ("incidence_angle_mid", np.float32),
        ("incidence_angle_aft", np.float32),
        ("azimuth_angle_for", np.float32),
        ("azimuth_angle_mid", np.float32),
        ("azimuth_angle_aft", np.float32),
        ("kp_for", np.float32),
        ("kp_mid", np.float32),
        ("kp_aft", np.float32),
        ("n_echos_for", np.int8),
        ("n_echos_mid", np.int8),
        ("n_echos_aft", np.int8),
        ("all_backscatter_for", np.float32),
        ("all_backscatter_mid", np.float32),
        ("all_backscatter_aft", np.float32),
        ("all_backscatter_std_for", np.float32),
        ("all_backscatter_std_mid", np.float32),
        ("all_backscatter_std_aft", np.float32),
        ("all_incidence_angle_for", np.float32),
        ("all_incidence_angle_mid", np.float32),
        ("all_incidence_angle_aft", np.float32),
        ("all_azimuth_angle_for", np.float32),
        ("all_azimuth_angle_mid", np.float32),
        ("all_azimuth_angle_aft", np.float32),
        ("all_kp_for", np.float32),
        ("all_kp_mid", np.float32),
        ("all_kp_aft", np.float32),
        ("all_n_echos_for", np.int8),
        ("all_n_echos_mid", np.int8),
        ("all_n_echos_aft", np.int8),
    ])

    @staticmethod
    def fn_read_fmt(timestamp):
        """
        Format a timestamp to search as YYYYMMDD*, for use in a regex
        that will match all files covering a single given date.

        Parameters
        ----------
        timestamp: datetime.datetime
            Timestamp to format

        Returns
        -------
        dict
            Dictionary of formatted strings
        """
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": "[ABC]",
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp):
        return {
            "satellite_folder": {
                "satellite": "metop_[abc]"
            },
            "year_folder": {
                "year": f"{timestamp.year}"
            },
        }

    def __init__(self, filename, **kwargs):
        super().__init__(filename, "netcdf4", **kwargs)


cell_io_catalog = {
    "H129": AscatH129Cell,
    "H129_V1.0": AscatH129v1Cell,
    "H121_V1.0": AscatH121v1Cell,
    "H122": AscatH122Cell,
    "SIG0_6.25": AscatSIG0Cell6250m,
    "SIG0_12.5": AscatSIG0Cell12500m,
}

swath_io_catalog = {
    "H129": AscatH129Swath,
    "H129_V1.0": AscatH129v1Swath,
    "H121_V1.0": AscatH121v1Swath,
    "H122": AscatH122Swath,
    "SIG0_6.25": AscatSIG0Swath6250m,
    "SIG0_12.5": AscatSIG0Swath12500m,
}

swath_fname_regex_lookup = {
    "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP[ABC]-6.25-H129_C_LIIB_.*_.*_.*____.nc":
        "H129",
    "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP[ABC]-6.25km-H129_C_LIIB_.*_.*_.*____.nc":
        "H129_V1.0",
    "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP[ABC]-12.5km-H121_C_LIIB_.*_.*_.*____.nc":
        "H121_V1.0",
    "ascat_ssm_nrt_6.25km_.*Z_.*Z_metop-[ABC]_h122.nc":
        "H122",
    "W_IT-HSAF-ROME,SAT,SIG0-ASCAT-METOP[ABC]-6.25_C_LIIB_.*_.*_.*____.nc":
        "SIG0_6.25",
    "W_IT-HSAF-ROME,SAT,SIG0-ASCAT-METOP[ABC]-12.5_C_LIIB_.*_.*_.*____.nc":
        "SIG0_12.5",
    "ascat_h129-v1.0_6.25km_.*_.*_.*.nc": "H129",
    "ascat_h121-v1.0_12.5km_.*_.*_.*.nc": "H121",
    "ascat_h122_6.25km_.*_.*_*.nc": "H122",
    "ascat_h129_6.25km_.*_.*_.*.nc": "H122"
}


def get_swath_product_id(filename):
    """
    Get product identifier from filename.

    Parameters
    ----------
    filename : str
        Filename.

    Returns
    -------
    product_id : str
        Product identifier.
    """
    product_id = None

    for pattern, swath_product_id in swath_fname_regex_lookup.items():
        if re.match(pattern, filename):
            product_id = swath_product_id

    return product_id



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Readers for H SAF soil moisture products.
"""

import os
import glob
import warnings
from datetime import datetime

import numpy as np

try:
    import pygrib
except ImportError:
    warnings.warn(
        'pygrib can not be imported GRIB files (H14) can not be read.')

from ascat.file_handling import ChronFiles
from ascat.file_handling import Filenames
from ascat.eumetsat.level2 import AscatL2File
from ascat.read_native.cdr import AscatGriddedNcTs


class AscatNrtBufrFileList(ChronFiles):
    """
    Class reading ASCAT NRT BUFR files.
    """

    def __init__(self,
                 root_path,
                 product_id='*',
                 filename_template=None,
                 subfolder_template=None):
        """
        Initialize.
        """
        if filename_template is None:
            filename_template = '{product_id}_{date}*.buf'

        self.product_id = product_id

        super().__init__(root_path,
                         AscatL2File,
                         filename_template,
                         sf_templ=subfolder_template)

    def _fmt(self, timestamp):
        """
        Definition of filename and subfolder format.

        Parameters
        ----------
        timestamp : datetime
            Time stamp.

        Returns
        -------
        fn_fmt : dict
            Filename format.
        sf_fmt : dict
            Subfolder format.
        """
        fn_read_fmt = {
            'date': timestamp.strftime('%Y%m%d_%H%M%S'),
            'product_id': self.product_id
        }
        sf_read_fmt = None
        fn_write_fmt = None
        sf_write_fmt = None

        return fn_read_fmt, sf_read_fmt, fn_write_fmt, sf_write_fmt

    def _parse_date(self, filename):
        """
        Parse date from filename.

        Parameters
        ----------
        filename : str
            Filename.

        Returns
        -------
        date : datetime
            Parsed date.
        """
        return datetime.strptime(
            os.path.basename(filename)[4:19], '%Y%m%d%_H%M%S')

    def _merge_data(self, data):
        """
        Merge data.

        Parameters
        ----------
        data : list
            List of array.

        Returns
        -------
        data : numpy.ndarray
            Data.
        """
        return np.hstack(data)


class H14Grib(Filenames):
    """
    Class reading H14 soil moisture in GRIB format.
    """

    def __init__(self,
                 filename,
                 expand_grid=True,
                 metadata_fields=['units', 'name']):
        """

        Parameters
        ----------
        expand_grid : boolean, optional
            if set the images will be expanded to a 2D image during reading
            if false the images will be returned as 1D arrays on the
            reduced gaussian grid
            Default: True
        metadata_fields: list, optional
            fields of the message to put into the metadata dictionary.
        """
        super().__init__(filename)
        self.expand_grid = expand_grid
        self.metadata_fields = metadata_fields
        self.pygrib1 = True

        if int(pygrib.__version__[0]) > 1:
            self.pygrib1 = False

    def _read(self, filename, timestamp=None):
        """
        Read specific image for given datetime timestamp.

        Parameters
        ----------
        timestamp : datetime.datetime
            exact observation timestamp of the image that should be read

        Returns
        -------
        data : dict
            dictionary of numpy arrays that hold the image data for each
            variable of the dataset
        """
        if self.pygrib1:
            param_names = {
                '40': 'SM_layer1_0-7cm',
                '41': 'SM_layer2_7-28cm',
                '42': 'SM_layer3_28-100cm',
                '43': 'SM_layer4_100-289cm'
            }
        else:
            param_names = {
                'SWI1 Soil wetness index in layer 1': 'SM_layer1_0-7cm',
                'SWI2 Soil wetness index in layer 2': 'SM_layer2_7-28cm',
                'SWI3 Soil wetness index in layer 3': 'SM_layer3_28-100cm',
                'SWI4 Soil wetness index in layer 4': 'SM_layer4_100-289cm',
                'Soil wetness index in layer 1': 'SM_layer1_0-7cm',
                'Soil wetness index in layer 2': 'SM_layer2_7-28cm',
                'Soil wetness index in layer 3': 'SM_layer3_28-100cm',
                'Soil wetness index in layer 4': 'SM_layer4_100-289cm'
            }
        data = {}
        metadata = {}

        with pygrib.open(filename) as grb:
            for i, message in enumerate(grb):
                message.expand_grid(self.expand_grid)
                if i == 1:
                    data['lat'], data['lon'] = message.latlons()

                data[param_names[message['parameterName']]] = message.values

                # read and store metadata
                md = {}
                for k in self.metadata_fields:
                    if message.valid_key(k):
                        md[k] = message[k]

                metadata[param_names[message['parameterName']]] = md

        return data


class H14GribFileList(ChronFiles):
    """
    Reads H SAF H08 data.
    """

    def __init__(self, path):
        """
        Initialize.
        """
        fn_templ = 'H14_{date}.grib'
        sf_templ = {'month': 'h14_{date}_grib'}

        super().__init__(path, H14Grib, fn_templ, sf_templ=sf_templ)

    def _fmt(self, timestamp):
        """
        Definition of filename and subfolder format.

        Parameters
        ----------
        timestamp : datetime
            Time stamp.

        Returns
        -------
        fn_fmt : dict
            Filename format.
        sf_fmt : dict
            Subfolder format.
        """
        fn_read_fmt = {'date': timestamp.strftime('%Y%m%d%H')}
        sf_read_fmt = {'month': {'date': timestamp.strftime('%Y%m')}}
        fn_write_fmt = None
        sf_write_fmt = None

        return fn_read_fmt, sf_read_fmt, fn_write_fmt, sf_write_fmt

    def _parse_date(self, filename):
        """
        Parse date from filename.

        Parameters
        ----------
        filename : str
            Filename.

        Returns
        -------
        date : datetime
            Parsed date.
        """
        return datetime.strptime(os.path.basename(filename)[4:15], '%Y%m%d%H')

    def read_period(dt_start, dt_end, delta):
        """
        Read period not implemented.
        """
        raise NotImplementedError()


class AscatSsmDataRecord(AscatGriddedNcTs):
    """
    Class reading Metop ASCAT soil moisture data record.
    """

    def __init__(self,
                 cdr_path,
                 grid_path,
                 fn_format=None,
                 grid_filename='TUW_WARP5_grid_info_2_2.nc',
                 static_layer_path=None,
                 **kwargs):
        """
        Initialize.

        Parameters
        ----------
        cdr_path : str
            Path to Climate Data Record (CDR) data set.
        grid_path : str
            Path to grid file.
        grid_filename : str
            Name of grid file.
        static_layer_path : str
            Path to static layer files.

        Attributes
        ----------
        grid : pygeogrids.CellGrid
            Cell grid.
        """
        if fn_format is None:
            first_file = glob.glob(os.path.join(cdr_path, '*.nc'))

            if len(first_file) == 0:
                raise RuntimeError('No files found')

            version = os.path.basename(first_file[0]).rsplit('_', 1)[0]
            fn_format = '{:}_{{:04d}}'.format(version)

        grid_filename = os.path.join(grid_path, grid_filename)

        super().__init__(cdr_path, fn_format, grid_filename, static_layer_path,
                         **kwargs)
